#!/usr/bin/env python
'''
file containing plotter functions for example scripts
'''
import os
import numpy as np
import h5py
import LFPy
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.collections import PolyCollection

def remove_axis_junk(ax, which=['right', 'top']):
    '''remove upper and right axis'''
    for loc, spine in ax.spines.items():
        if loc in which:
            spine.set_color('none')            
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')


def plot_signal_sum(ax, z, fname='LFPsum.h5', unit='mV',
                    ylabels=True, scalebar=True, vlimround=None,
                    T=[0, 1000], color='k',
                    label=''):
    '''
    on axes plot the signal contributions
    
    args:
    ::
        ax : matplotlib.axes.AxesSubplot object
        z : np.ndarray
        T : list, [tstart, tstop], which timeinterval
        ylims : list, set range of yaxis to scale with other plots
        fancy : bool, 
        scaling_factor : float, scaling factor (e.g. to scale 10% data set up)
    '''    
    #open file and get data, samplingrate
    f = h5py.File(fname)
    data = f['data'].value
    dataT = data.T - data.mean(axis=1)
    data = dataT.T
    srate = f['srate'].value
    
    #close file object
    f.close()

    # normalize data for plot
    tvec = np.arange(data.shape[1]) * 1000. / srate
    slica = (tvec <= T[1]) & (tvec >= T[0])
    zvec = np.r_[z]
    zvec = np.r_[zvec, zvec[-1] + np.diff(zvec)[-1]]
    vlim = abs(data[:, slica]).max()
    if vlimround is None:
        vlimround = 2.**np.round(np.log2(vlim))
    yticklabels=[]
    yticks = []
    
    colors = [color]*data.shape[0]
    
    for i, z in enumerate(z):
        if i == 0:
            ax.plot(tvec[slica], data[i, slica] * 100 / vlimround + z,
                    color=colors[i], rasterized=False, label=label,
                    clip_on=False)
        else: 
            ax.plot(tvec[slica], data[i, slica] * 100 / vlimround + z,
                    color=colors[i], rasterized=False, clip_on=False)
        yticklabels.append('ch. %i' % (i+1))
        yticks.append(z)
     
    if scalebar:
        ax.plot([tvec[slica][-1], tvec[slica][-1]],
                [-0, -100], lw=2, color='k', clip_on=False)
        ax.text(tvec[slica][-1]+np.diff(T)*0.02, -50,
                r'%g %s' % (vlimround, unit),
                color='k', rotation='vertical')

    ax.axis(ax.axis('tight'))
    ax.yaxis.set_ticks(yticks)
    if ylabels:
        ax.yaxis.set_ticklabels(yticklabels)
    else:
        ax.yaxis.set_ticklabels([])

    for loc, spine in ax.spines.items():
        if loc in ['right', 'top']:
            spine.set_color('none')            
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')
    ax.set_xlabel(r'time (ms)', labelpad=0)


def plot_pop_scatter(ax, somapos, isometricangle, marker, color):
    #scatter plot setting appropriate zorder for each datapoint by binning
    for lower in np.arange(-600, 601, 20):
        upper = lower + 20
        inds = (somapos[:, 1] >= lower) & (somapos[:, 1] < upper)
        if np.any(inds):
            ax.scatter(somapos[inds, 0],
                       somapos[inds, 2] - somapos[inds, 1] *
                       np.sin(isometricangle),
                   s=30, facecolors=color, edgecolors='gray', linewidth=0.1,
                   zorder=lower,
                   marker = marker, clip_on=False, rasterized=True)


def plot_population(ax,
                    populationParams,
                    electrodeParams,
                    layerBoundaries,
                    aspect='equal',
                    isometricangle=np.pi/12,
                    X=['EX', 'IN'],
                    markers=['^', 'o'],
                    colors=['r', 'b'],
                    layers = ['upper', 'lower'],
                    title='positions'):
    '''
    Plot the geometry of the column model, optionally with somatic locations
    and optionally with reconstructed neurons
    
    kwargs:
    ::
        ax : matplotlib.axes.AxesSubplot
        aspect : str
            matplotlib.axis argument
        isometricangle : float
            pseudo-3d view angle
        plot_somas : bool
            plot soma locations
        plot_morphos : bool
            plot full morphologies
        num_unitsE : int
            number of excitatory morphos plotted per population
        num_unitsI : int
            number of inhibitory morphos plotted per population
        clip_dendrites : bool
            draw dendrites outside of axis
        mainpops : bool
            if True, plot only main pops, e.g. b23 and nb23 as L23I
    
    return:
    ::
        axis : list
            the plt.axis() corresponding to input aspect
    '''

    remove_axis_junk(ax, ['right', 'bottom', 'left', 'top'])

    
    # DRAW OUTLINE OF POPULATIONS 
    ax.xaxis.set_ticks([])
    ax.yaxis.set_ticks([])

    #contact points
    ax.plot(electrodeParams['x'],
            electrodeParams['z'],
            '.', marker='o', markersize=5, color='k', zorder=0)

    #outline of electrode       
    x_0 = np.array(electrodeParams['r_z'])[1, 1:-1]
    z_0 = np.array(electrodeParams['r_z'])[0, 1:-1]
    x = np.r_[x_0[-1], x_0[::-1], -x_0[1:], -x_0[-1]]
    z = np.r_[100, z_0[::-1], z_0[1:], 100]
    ax.fill(x, z, color=(0.5, 0.5, 0.5), lw=None, zorder=-0.1)

    #outline of populations:
    #fetch the population radius from some population
    r = populationParams[populationParams.keys()[0]]['radius']

    theta0 = np.linspace(0, np.pi, 20)
    theta1 = np.linspace(np.pi, 2*np.pi, 20)
    
    zpos = np.r_[np.array(layerBoundaries)[:, 0],
                 np.array(layerBoundaries)[-1, 1]]
    
    for i, z in enumerate(np.mean(layerBoundaries, axis=1)):
        ax.text(r, z, ' %s' % layers[i],
                va='center', ha='left', rotation='vertical')

    for i, zval in enumerate(zpos):
        if i == 0:
            ax.plot(r*np.cos(theta0),
                    r*np.sin(theta0)*np.sin(isometricangle)+zval,
                    color='k', zorder=-r, clip_on=False)
            ax.plot(r*np.cos(theta1),
                    r*np.sin(theta1)*np.sin(isometricangle)+zval,
                    color='k', zorder=r, clip_on=False)
        else:
            ax.plot(r*np.cos(theta0),
                    r*np.sin(theta0)*np.sin(isometricangle)+zval,
                    color='gray', zorder=-r, clip_on=False)
            ax.plot(r*np.cos(theta1),
                    r*np.sin(theta1)*np.sin(isometricangle)+zval,
                    color='k', zorder=r, clip_on=False)
    
    ax.plot([-r, -r], [zpos[0], zpos[-1]], 'k', zorder=0, clip_on=False)
    ax.plot([r, r], [zpos[0], zpos[-1]], 'k', zorder=0, clip_on=False)
    
    #plot a horizontal radius scalebar
    ax.plot([0, r], [z_0.min()]*2, 'k', lw=2, zorder=0, clip_on=False)
    ax.text(r / 2., z_0.min()-100, 'r = %i $\mu$m' % int(r), ha='center')
    
    #plot a vertical depth scalebar
    ax.plot([-r]*2, [z_0.min()+50, z_0.min()-50],
        'k', lw=2, zorder=0, clip_on=False)
    ax.text(-r, z_0.min(), r'100 $\mu$m', va='center', ha='right')
    
    ax.set_yticks([])
    ax.set_yticklabels([])
    
    #fake ticks:
    for pos in zpos:
        ax.text(-r, pos, 'z=%i-' % int(pos), ha='right', va='center')
 
    ax.set_title(title)
   
    axis = ax.axis(ax.axis(aspect))


def plot_soma_locations(ax, X, populations_path, isometricangle, markers, colors):
    for (pop, marker, color) in zip(X, markers, colors):
        #get the somapos
        fname = os.path.join(populations_path,
                             '%s_population_somapos.gdf' % pop)
        
        somapos = np.loadtxt(fname).reshape((-1, 3))
    
        plot_pop_scatter(ax, somapos, isometricangle, marker, color)
        

def plot_morphologies(ax, X, isometricangle, markers, colors,
                      populations_path, cellParams, fraction=1):
    for (pop, marker, color) in zip(X, markers, colors):
        #get the somapos
        fname = os.path.join(populations_path,
                             '%s_population_somapos.gdf' % pop)
        
        somapos = np.loadtxt(fname).reshape((-1, 3))
        n = somapos.shape[0]
        
        rotations = [{} for x in range(n)]
        fname = os.path.join(populations_path,
                             '%s_population_rotations.h5' % pop)
        f = h5py.File(fname, 'r')
        
        for key, value in list(f.items()):
            for i, rot in enumerate(value.value):
                rotations[i].update({key : rot})
        
        
        #plot some units
        for j in range(int(n*fraction)):
            cell = LFPy.Cell(morphology=cellParams[pop]['morphology'],
                             nsegs_method='lambda_f',
                             lambda_f=100,
                             extracellular=False
                            )
            cell.set_pos(somapos[j, 0], somapos[j, 1], somapos[j, 2])
            cell.set_rotation(**rotations[j])

            #set up a polycollection
            zips = []
            for x, z in cell.get_idx_polygons():
                zips.append(list(zip(x, z-somapos[j, 1] * np.sin(isometricangle)
                                     )))
            
            polycol = PolyCollection(zips,
                                     edgecolors='gray',
                                     facecolors=color,
                                     linewidths=(0.1),
                                     zorder=somapos[j, 1],
                                     clip_on=False,
                                     rasterized=False)

            ax.add_collection(polycol)
            


def plot_individual_morphologies(ax, X, isometricangle, markers, colors,
                                 cellParams, populationParams):
    
    depth = []
    offsets = []
    depth0 = np.inf
    offset = -75
    for i, y in enumerate(X):
        d = populationParams[y]
        depth += [(d['z_min']+d['z_max'])/2]
        if depth0 == depth[-1]:
            offset += 150
        else:
            offset = -75 - i*20
        depth0 = depth[-1]
        offsets += [offset]
    somapos = np.c_[offsets, np.zeros(len(offsets)), depth]
    
    for i, (pop, marker, color) in enumerate(zip(X, markers, colors)):
        
        cell = LFPy.Cell(morphology=cellParams[pop]['morphology'],
                         nsegs_method='lambda_f',
                         lambda_f=100,
                         extracellular=False
                        )
        cell.set_pos(somapos[i, 0], somapos[i, 1], somapos[i, 2])
        
        #set up a polycollection
        zips = []
        for x, z in cell.get_idx_polygons():
            zips.append(list(zip(x, z-somapos[i, 1] * np.sin(isometricangle))))
        
        polycol = PolyCollection(zips,
                                 edgecolors='gray',
                                 facecolors=color,
                                 linewidths=(0.1),
                                 zorder=somapos[i, 1],
                                 clip_on=False,
                                 rasterized=False)

        ax.add_collection(polycol)


def normalize(x):
    '''normalize x to have mean 0 and unity standard deviation'''
    x = x.astype(float)
    x -= x.mean()
    return x / float(x.std())


def plot_correlation(z_vec, x0, x1, ax, lag=20., title='firing_rate vs LFP'):
    ''' mls
    on axes plot the correlation between x0 and x1
    
    args:
    ::
        x0 : first dataset
        x1 : second dataset - e.g., the multichannel LFP
        ax : matplotlib.axes.AxesSubplot object
        title : text to be used as current axis object title
    '''
    zvec = np.r_[z_vec]
    zvec = np.r_[zvec, zvec[-1] + np.diff(zvec)[-1]]
    
    xcorr_all=np.zeros((np.size(z_vec), x0.shape[0]))
    for i, z in enumerate(z_vec):
        x2 = x1[i, ]
        xcorr1 = np.correlate(normalize(x0),
                              normalize(x2), 'same') / x0.size
        xcorr_all[i,:]=xcorr1  

    # Find limits for the plot
    vlim = abs(xcorr_all).max()
    vlimround = 2.**np.round(np.log2(vlim))
    
    yticklabels=[]
    yticks = []
    ylimfound=np.zeros((1,2))
    for i, z in enumerate(z_vec):
        ind = np.arange(x0.size) - x0.size/2
        ax.plot(ind, xcorr_all[i,::-1] * 100. / vlimround + z, 'k',
                clip_on=True, rasterized=False)
        yticklabels.append('ch. %i' %(i+1))
        yticks.append(z)    

    remove_axis_junk(ax)
    ax.set_title(title)
    ax.set_xlabel(r'lag $\tau$ (ms)')

    ax.set_xlim(-lag, lag)
    ax.set_ylim(z-100, 100)
    
    axis = ax.axis()
    ax.vlines(0, axis[2], axis[3], 'r', 'dotted')

    ax.yaxis.set_ticks(yticks)
    ax.yaxis.set_ticklabels(yticklabels)

    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')

    # Create a scaling bar
    ax.plot([lag, lag],
        [0, 100], lw=2, color='k', clip_on=False)
    ax.text(lag, 50, r'CC=%.2f' % vlimround,
            rotation='vertical', va='center')



#!/usr/bin/env python
'''
Hybrid LFP scheme example script, applying the methodology with the model of:

Potjans, T. and Diesmann, M. "The Cell-Type Specific Cortical Microcircuit:
Relating Structure and Activity in a Full-Scale Spiking Network Model".
Cereb. Cortex (2014) 24 (3): 785-806.
doi: 10.1093/cercor/bhs358

Synopsis of the main simulation procedure:
1. Loading of parameterset
    a. network parameters
    b. parameters for hybrid scheme
2. Set up file destinations for different simulation output
3. network simulation
    a. execute network simulation using NEST (www.nest-simulator.org)
    b. merge nest spike output from different MPI ranks
4. Create a object-representation that uses sqlite3 of all the spiking output 
5. Iterate over post-synaptic populations:
    a. Create Population object with appropriate parameters for
       each specific population
    b. Run all computations for populations
    c. Postprocess simulation output of all cells in population
6. Postprocess all cell- and population-specific output data
7. Create a tarball for all non-redundant simulation output

The full simulation can be evoked by issuing a mpirun call, such as
    mpirun -np 64 python example_microcircuit.py
where the number 64 is the desired number of MPI threads & CPU cores

Given the size of the network and demands for the multi-compartment LFP-
predictions using the present scheme, running the model on a large scale
compute facility is strongly encouraged.

'''
import os
import numpy as np
from time import time
import nest
from hybridLFPy import PostProcess, Population, CachedNetwork
from hybridLFPy import setup_file_dest, helpers
from glob import glob
import neuron
from mpi4py import MPI


#set some seed values
SEED = 12345678
SIMULATIONSEED = 12345678
np.random.seed(SEED)


################# Initialization of MPI stuff ##################################
COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()

#if True, execute full model. If False, do only the plotting. Simulation results
#must exist.
properrun = True


#check if mod file for synapse model specified in expisyn.mod is loaded
if not hasattr(neuron.h, 'ExpSynI'):
    if RANK == 0:
        os.system('nrnivmodl')
    COMM.Barrier()
    neuron.load_mechanisms('.')
    

################################################################################
## PARAMETERS
################################################################################

from example_microcircuit_params import multicompartment_params, \
                                point_neuron_network_params

#Full set of parameters including network parameters
params = multicompartment_params()


################################################################################
# Function declaration(s)
################################################################################

def merge_gdf(model_params, raw_label='spikes_', file_type='gdf',
              fileprefix='spikes'):
    '''
    NEST produces one file per virtual process containing recorder output. 
    This function gathers and combines them into one single file per 
    network population.
    
    Parameters
    ----------
    model_params : object
        network parameters object
    
    Returns
    -------
    None
    
    '''
    def get_raw_gids(model_params):
        '''
        Reads text file containing gids of neuron populations as created within
        the NEST simulation. These gids are not continuous as in the simulation
        devices get created in between.
        
        Parameters
        ----------
        model_params : object
            network parameters object
        
        
        Returns
        -------
        gids : list
            list of neuron ids and value (spike time, voltage etc.)
        
        '''
        gidfile = open(os.path.join(model_params.raw_nest_output_path,
                                    model_params.GID_filename),'r') 
        gids = [] 
        for l in gidfile :
            a = l.split()
            gids.append([int(a[0]),int(a[1])])
        return gids
    
    #some preprocessing
    raw_gids = get_raw_gids(model_params)
    pop_sizes = [raw_gids[i][1]-raw_gids[i][0]+1
                 for i in np.arange(model_params.Npops)]
    raw_first_gids =  [raw_gids[i][0] for i in np.arange(model_params.Npops)]
    converted_first_gids = [int(1 + np.sum(pop_sizes[:i]))
                            for i in np.arange(model_params.Npops)]

    for pop_idx in np.arange(model_params.Npops):
        if pop_idx % SIZE == RANK:
            files = glob(os.path.join(model_params.raw_nest_output_path,
                                      raw_label + str(pop_idx) +
                                      '*.' + file_type))
            gdf = [] # init
            for f in files:
                new_gdf = helpers.read_gdf(f)
                for line in new_gdf:
                    line[0] = line[0] - raw_first_gids[pop_idx] + \
                              converted_first_gids[pop_idx]
                    gdf.append(line)
            
            print 'writing: %s' % os.path.join(model_params.spike_output_path,
                                            fileprefix +
                                            '_%s.gdf' % model_params.X[pop_idx])
            helpers.write_gdf(gdf, os.path.join(model_params.spike_output_path,
                                        fileprefix +
                                        '_%s.gdf' % model_params.X[pop_idx]))
    
    COMM.Barrier()

    return


def dict_of_numpyarray_to_dict_of_list(d):
    '''
    Convert dictionary containing numpy arrays to dictionary containing lists
    
    Parameters
    ----------
    d : dict
        sli parameter name and value as dictionary key and value pairs
    
    Returns
    -------
    d : dict
        modified dictionary
    
    '''
    for key,value in d.iteritems():
        if isinstance(value,dict):  # if value == dict 
            # recurse
            d[key] = dict_of_numpyarray_to_dict_of_list(value)
        elif isinstance(value,np.ndarray): # or isinstance(value,list) :
            d[key] = value.tolist()
    return d


def send_nest_params_to_sli(p):
    '''
    Read parameters and send them to SLI
    
    Parameters
    ----------
    p : dict
        sli parameter name and value as dictionary key and value pairs
    
    Returns
    -------
    None
    '''
    for name in p.keys():
        value = p[name]
        if type(value) == np.ndarray:
            value = value.tolist()
        if type(value) == dict:
            value = dict_of_numpyarray_to_dict_of_list(value)
        if name == 'neuron_model': # special case as neuron_model is a
                                   # NEST model and not a string
            try:
                nest.sli_run('/'+name)
                nest.sli_push(value)
                nest.sli_run('eval')
                nest.sli_run('def')
            except: 
                print 'Could not put variable %s on SLI stack' % (name)
                print type(value)
        else:
            try:
                nest.sli_run('/'+name)
                nest.sli_push(value)
                nest.sli_run('def')
            except: 
                print 'Could not put variable %s on SLI stack' % (name)
                print type(value)
    return


def sli_run(parameters=object(),
            fname='microcircuit.sli',
            verbosity='M_ERROR'):
    '''
    Takes parameter-class and name of main sli-script as input, initiating the
    simulation.
    
    Parameters
    ----------
    parameters : object
        parameter class instance
    fname : str
        path to sli codes to be executed
    verbosity : str,
        nest verbosity flag
    
    Returns
    -------
    None
    
    '''
    # Load parameters from params file, and pass them to nest
    # Python -> SLI
    send_nest_params_to_sli(vars(parameters))
    
    #set SLI verbosity
    nest.sli_run("%s setverbosity" % verbosity)
    
    # Run NEST/SLI simulation
    nest.sli_run('(%s) run' % fname)


###############################################################################
# MAIN simulation procedure
###############################################################################

#tic toc
tic = time()

if properrun:
    #set up the file destination
    setup_file_dest(params, clearDestination=True)

######## Perform network simulation ############################################

if properrun:
    #initiate nest simulation with only the point neuron network parameter class
    networkParams = point_neuron_network_params()
    sli_run(parameters=networkParams,
            fname='microcircuit.sli',
            verbosity='M_WARNING')
    
    #preprocess the gdf files containing spiking output, voltages, weighted and
    #spatial input spikes and currents:
    merge_gdf(networkParams,
              raw_label=networkParams.spike_detector_label,
              file_type='gdf',
              fileprefix=params.networkSimParams['label'])


#Create an object representation of the simulation output that uses sqlite3
networkSim = CachedNetwork(**params.networkSimParams)


toc = time() - tic
print 'NEST simulation and gdf file processing done in  %.3f seconds' % toc


####### Set up populations #####################################################

if properrun:
    #iterate over each cell type, run single-cell simulations and create
    #population object
    for i, y in enumerate(params.y):
        #create population:
        pop = Population(
                #parent class parameters
                cellParams = params.yCellParams[y],
                rand_rot_axis = params.rand_rot_axis[y],
                simulationParams = params.simulationParams,
                populationParams = params.populationParams[y],
                y = y,
                layerBoundaries = params.layerBoundaries,
                electrodeParams = params.electrodeParams,
                savelist = params.savelist,
                savefolder = params.savefolder,
                calculateCSD = params.calculateCSD,
                dt_output = params.dt_output, 
                POPULATIONSEED = SIMULATIONSEED + i,
                #daughter class kwargs
                X = params.X,
                networkSim = networkSim,
                k_yXL = params.k_yXL[y],
                synParams = params.synParams[y],
                synDelayLoc = params.synDelayLoc[y],
                synDelayScale = params.synDelayScale[y],
                J_yX = params.J_yX[y],
                recordSingleContribFrac = params.recordSingleContribFrac,
            )
    
        #run population simulation and collect the data
        pop.run()
        pop.collect_data()
    
        #object no longer needed
        del pop


####### Postprocess the simulation output ######################################


#reset seed, but output should be deterministic from now on
np.random.seed(SIMULATIONSEED)

if properrun:
    #do some postprocessing on the collected data, i.e., superposition
    #of population LFPs, CSDs etc
    postproc = PostProcess(y = params.y,
                           dt_output = params.dt_output,
                           savefolder = params.savefolder,
                           mapping_Yy = params.mapping_Yy,
                           )
    
    #run through the procedure
    postproc.run()
    
    #create tar-archive with output
    postproc.create_tar_archive()

#tic toc
print 'Execution time: %.3f seconds' %  (time() - tic)


################################################################################
# Create set of plots from simulation output
################################################################################

########## matplotlib settings #################################################
import matplotlib.pyplot as plt
from example_plotting import *

plt.close('all')

#turn off interactive plotting
plt.ioff()

if RANK == 0:
    #create network raster plot
    x, y = networkSim.get_xy((500, 1000), fraction=1)
    fig, ax = plt.subplots(1, figsize=(5,8))
    fig.subplots_adjust(left=0.2)
    networkSim.plot_raster(ax, (500, 1000), x, y, markersize=1, marker='o',
                           alpha=.5,legend=False, pop_names=True)
    remove_axis_junk(ax)
    ax.set_xlabel(r'$t$ (ms)', labelpad=0.1)
    ax.set_ylabel('population', labelpad=0.1)
    ax.set_title('network raster')
    fig.savefig(os.path.join(params.figures_path, 'network_raster.pdf'),
                dpi=300)
    #raise Exception
    
    #plot cell locations
    fig, ax = plt.subplots(1,1, figsize=(5,8))
    fig.subplots_adjust(left=0.2)
    plot_population(ax, params.populationParams, params.electrodeParams,
                    params.layerBoundaries,
                    X=params.y,
                    markers=['*' if 'b' in y else '^' for y in params.y],
                    colors=['b' if 'b' in y else 'r' for y in params.y],
                    layers = ['L1', 'L2/3', 'L4', 'L5', 'L6'],
                    isometricangle=np.pi/24, aspect='equal')
    ax.set_title('layers')
    fig.savefig(os.path.join(params.figures_path, 'layers.pdf'), dpi=300)
    

    #plot cell locations
    fig, ax = plt.subplots(1,1, figsize=(5,8))
    fig.subplots_adjust(left=0.2)
    plot_population(ax, params.populationParams, params.electrodeParams,
                    params.layerBoundaries,
                    X=params.y,
                    markers=['*' if 'b' in y else '^' for y in params.y],
                    colors=['b' if 'b' in y else 'r' for y in params.y],
                    layers = ['L1', 'L2/3', 'L4', 'L5', 'L6'],
                    isometricangle=np.pi/24, aspect='equal')
    plot_soma_locations(ax, X=params.y,
                        populations_path=params.populations_path,
                        markers=['*' if 'b' in y else '^' for y in params.y],
                        colors=['b' if 'b' in y else 'r' for y in params.y],
                        isometricangle=np.pi/24, )
    ax.set_title('soma positions')
    fig.savefig(os.path.join(params.figures_path, 'soma_locations.pdf'),
                dpi=150)
    
    
    #plot morphologies in their respective locations
    fig, ax = plt.subplots(1,1, figsize=(5,8))
    fig.subplots_adjust(left=0.2)
    plot_population(ax, params.populationParams, params.electrodeParams,
                    params.layerBoundaries,
                    X=params.y,
                    markers=['*' if 'b' in y else '^' for y in params.y],
                    colors=['b' if 'b' in y else 'r' for y in params.y],
                    layers = ['L1', 'L2/3', 'L4', 'L5', 'L6'],
                    isometricangle=np.pi/24, aspect='equal')
    plot_morphologies(ax,
                      X=params.y,
                      markers=['*' if 'b' in y else '^' for y in params.y],
                      colors=['b' if 'b' in y else 'r' for y in params.y],
                      isometricangle=np.pi/24,
                      populations_path=params.populations_path,
                      cellParams=params.yCellParams,
                      fraction=0.02)
    ax.set_title('LFP generators')
    fig.savefig(os.path.join(params.figures_path, 'populations.pdf'), dpi=300)


    #plot morphologies in their respective locations
    fig, ax = plt.subplots(1,1, figsize=(5,8))
    fig.subplots_adjust(left=0.2)
    plot_population(ax, params.populationParams, params.electrodeParams,
                    params.layerBoundaries,
                    X=params.y,
                    markers=['*' if 'b' in y else '^' for y in params.y],
                    colors=['b' if 'b' in y else 'r' for y in params.y],
                    layers = ['L1', 'L2/3', 'L4', 'L5', 'L6'],
                    isometricangle=np.pi/24, aspect='equal')
    plot_individual_morphologies(ax,
                   X=params.y,
                   markers=['*' if 'b' in y else '^' for y in params.y],
                   colors=['b' if 'b' in y else 'r' for y in params.y],
                   isometricangle=np.pi/24, cellParams=params.yCellParams,
                   populationParams=params.populationParams)
    ax.set_title('morphologies')
    fig.savefig(os.path.join(params.figures_path, 'cell_models.pdf'), dpi=300)


    #plot compound LFP and CSD traces
    fig = plt.figure(figsize=(13, 8))
    fig.subplots_adjust(left=0.075, right=0.95, bottom=0.075, top=0.95,
                        hspace=0.2, wspace=0.2)
    gs = gridspec.GridSpec(2,2)
    
    ax0 = fig.add_subplot(gs[:,0])
    ax1 = fig.add_subplot(gs[0, 1])
    ax2 = fig.add_subplot(gs[1, 1])
    ax0.set_title('network raster')
    ax1.set_title('CSD')
    ax2.set_title('LFP')

    T=(500, 700)
    
    x, y = networkSim.get_xy(T, fraction=1)
    networkSim.plot_raster(ax0, T, x, y, markersize=1, marker='o',
                           alpha=.5,legend=False, pop_names=True)
    remove_axis_junk(ax0)
    ax0.set_xlabel(r'$t$ (ms)', labelpad=0.1)
    ax0.set_ylabel('population', labelpad=0.1)
    

    plot_signal_sum(ax1, z=params.electrodeParams['z'],
                    fname=os.path.join(params.savefolder, 'CSDsum.h5'),
                    unit='$\mu$Amm$^{-3}$', T=T)
    ax1.set_xticklabels([])
    ax1.set_xlabel('')
    
    plot_signal_sum(ax2, z=params.electrodeParams['z'],
                    fname=os.path.join(params.savefolder, 'LFPsum.h5'),
                    unit='mV', T=T)
    ax2.set_xlabel('$t$ (ms)')
    
    fig.savefig(os.path.join(params.figures_path, 'compound_signals.pdf'),
                dpi=300)

    plt.show()




#!/usr/bin/env python
'''Tests for hybridLFPy classes and methods'''
import os
import sys
import numpy as np
import numpy.testing as nt
import unittest
import hybridLFPy


class TestHybridLFPy(unittest.TestCase):
    '''
    hybridLFPy test class
    
    To run tests, simply issue
    >>>import hybridLFPy
    >>>hybridLFPy.test()
    
    '''
    
    def __init__(self, *args, **kwargs):
        super(TestHybridLFPy, self).__init__(*args, **kwargs)
    
        self.networkSim = hybridLFPy.CachedNetwork(
            simtime=1000.,
            dt=0.1,
            spike_output_path=os.path.join(hybridLFPy.__path__[0], 'testing'),
            label='testing',
            ext='gdf',
            GIDs={'X' : [1, 100]}   
        )

    def test_CachedNetwork_01(self):
        '''test CachedNetwork 01'''
        assert(self.networkSim.X == ['X'])

    def test_CachedNetwork_02(self):
        '''test CachedNetwork 02'''
        nt.assert_array_equal(self.networkSim.N_X, np.array([100]))

    def test_CachedNetwork_03(self):
        '''test CachedNetwork 01'''
        nt.assert_array_equal(self.networkSim.nodes['X'], np.arange(100)+1)

    def test_CachedNetwork_04(self):
        '''test CachedNetwork 04''' 
        nt.assert_array_equal(self.networkSim.dbs['X'].neurons(), np.array([1, 50]))

    def test_CachedNetwork_05(self):
        '''test CachedNetwork 05'''
        nt.assert_array_equal(self.networkSim.dbs['X'].interval((0, 1000)), [(50, 100.0), (1, 200.0)])

    def test_CachedNetwork_06(self):
        '''test CachedNetwork 06'''
        nt.assert_array_equal(self.networkSim.dbs['X'].interval((0, 100)), [(50, 100.0)])

    def test_CachedNetwork_07(self):
        '''test CachedNetwork 07'''
        nt.assert_array_equal(self.networkSim.dbs['X'].interval((200, 1000)), [(1, 200.0)])
        
    def test_CachedNetwork_08(self):
        '''test CachedNetwork 08'''
        nt.assert_array_equal(self.networkSim.dbs['X'].select_neurons_interval(self.networkSim.dbs['X'].neurons(), (0,1000)), [np.array([ 200.]), np.array([ 100.])])


def test(verbosity=2):
    '''
    Run unittests for the CSD toolbox
    
    
    Arguments
    ---------
    verbosity : int
        verbosity level
    
    
    Returns
    -------
    None
    '''
    suite = unittest.TestLoader().loadTestsFromTestCase(TestHybridLFPy)
    unittest.TextTestRunner(verbosity=verbosity).run(suite)
    


if __name__ == '__main__':
    test()


#!/usr/bin/env python
'''
Hybrid LFP scheme example script, applying the methodology with a model
implementation similar to:

Nicolas Brunel. "Dynamics of Sparsely Connected Networks of Excitatory and
Inhibitory Spiking Neurons". J Comput Neurosci, May 2000, Volume 8,
Issue 3, pp 183-208


Synopsis of the main simulation procedure:
1. Loading of parameterset
    a. network parameters
    b. parameters for hybrid scheme
2. Set up file destinations for different simulation output
3. network simulation
    a. execute network simulation using NEST (www.nest-initiative.org)
    b. merge network output (spikes, currents, voltages)
4. Create a object-representation that uses sqlite3 of all the spiking output 
5. Iterate over post-synaptic populations:
    a. Create Population object with appropriate parameters for
       each specific population
    b. Run all computations for populations
    c. Postprocess simulation output of all cells in population
6. Postprocess all cell- and population-specific output data
7. Create a tarball for all non-redundant simulation output

The full simulation can be evoked by issuing a mpirun call, such as
mpirun -np 4 python example_brunel.py

Not recommended, but running it serially should also work, e.g., calling
python example_brunel.py


Given the size of the network and demands for the multi-compartment LFP-
predictions using the present scheme, running the model on nothing but a large-
scale compute facility is strongly discouraged.

'''
import os
import numpy as np
if 'DISPLAY' not in os.environ:
    import matplotlib
    matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib import gridspec
from time import time
import nest #not used, but load order determine if network is run in parallel
from hybridLFPy import PostProcess, Population, CachedNetwork, setup_file_dest
from NeuroTools.parameters import ParameterSet
import h5py
import neuron
from mpi4py import MPI

########## matplotlib settings #################################################
plt.close('all')
plt.rcParams.update({'figure.figsize': [10.0, 8.0]})


#set some seed values
SEED = 12345678
SIMULATIONSEED = 12345678
np.random.seed(SEED)


################# Initialization of MPI stuff ##################################
COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()

#if True, execute full model. If False, do only the plotting. Simulation results
#must exist.
properrun = True


#check if mod file for synapse model specified in alphaisyn.mod is loaded
if not hasattr(neuron.h, 'AlphaISyn'):
    if RANK == 0:
        os.system('nrnivmodl')
    COMM.Barrier()
    neuron.load_mechanisms('.')


################################################################################
## PARAMETERS
################################################################################

#Set up parameters using the NeuroTools.parameters.ParameterSet class.

#Access parameters defined in example script implementing the network using
#pynest, brunel_alpha_nest.py, adapted from the NEST v2.4.1 release. This will
#not execute the network model, but see below.
import brunel_alpha_nest as BN


#set up file destinations differentiating between certain output
PS = ParameterSet(dict(
    #Main folder of simulation output
    savefolder = 'simulation_output_example_brunel',
    
    #make a local copy of main files used in simulations
    sim_scripts_path = os.path.join('simulation_output_example_brunel',
                                    'sim_scripts'),
    
    #destination of single-cell output during simulation
    cells_path = os.path.join('simulation_output_example_brunel', 'cells'),
    
    #destination of cell- and population-specific signals, i.e., compund LFPs,
    #CSDs etc.
    populations_path = os.path.join('simulation_output_example_brunel',
                                    'populations'),
    
    #location of spike output from the network model
    spike_output_path = BN.spike_output_path,
    
    #destination of figure file output generated during model execution
    figures_path = os.path.join('simulation_output_example_brunel', 'figures')
))


#population (and cell type) specific parameters
PS.update(dict(
    #no cell type specificity within each E-I population
    #hence X == x and Y == X
    X = ["EX", "IN"],
    
    #population-specific LFPy.Cell parameters
    cellParams = dict(
        #excitory cells
        EX = dict(
            morphology = 'morphologies/ex.hoc',
            v_init = BN.neuron_params['E_L'],
            rm = BN.neuron_params['tau_m'] * 1E3 / 1.0, #assume cm=1.
            cm = 1.0,
            Ra = 150,
            e_pas = BN.neuron_params['E_L'],    
            nsegs_method = 'lambda_f',
            lambda_f = 100,
            timeres_NEURON = BN.dt,
            timeres_python = BN.dt,
            tstartms = 0,
            tstopms = BN.simtime,
            verbose = False,        
        ),
        #inhibitory cells
        IN = dict(
            morphology = 'morphologies/in.hoc',
            v_init = BN.neuron_params['E_L'],
            rm = BN.neuron_params['tau_m'] * 1E3 / 1.0, #assume cm=1.
            cm = 1.0,
            Ra = 150,
            e_pas = BN.neuron_params['E_L'],    
            nsegs_method = 'lambda_f',
            lambda_f = 100,
            timeres_NEURON = BN.dt,
            timeres_python = BN.dt,
            tstartms = 0,
            tstopms = BN.simtime,
            verbose = False,                    
    )),
    
    #assuming excitatory cells are pyramidal
    rand_rot_axis = dict(
        EX = ['z'],
        IN = ['x', 'y', 'z'],
    ),
    
    
    #kwargs passed to LFPy.Cell.simulate()
    simulationParams = dict(),
    
    #set up parameters corresponding to cylindrical model populations
    populationParams = dict(
        EX = dict(
            number = BN.NE,
            radius = 100,
            z_min = -450,
            z_max = -350,
            min_cell_interdist = 1.,            
            ),
        IN = dict(
            number = BN.NI,
            radius = 100,
            z_min = -450,
            z_max = -350,
            min_cell_interdist = 1.,            
            ),
    ),
    
    #set the boundaries between the "upper" and "lower" layer
    layerBoundaries = [[0., -300],
                      [-300, -500]],
    
    #set the geometry of the virtual recording device
    electrodeParams = dict(
            #contact locations:
            x = [0]*6,
            y = [0]*6,
            z = [x*-100. for x in range(6)],
            #extracellular conductivity:
            sigma = 0.3,
            #contact surface normals, radius, n-point averaging
            N = [[1, 0, 0]]*6,
            r = 5,
            n = 20,
            seedvalue = None,
            #dendrite line sources, soma as sphere source (Linden2014)
            method = 'som_as_point',
            #no somas within the constraints of the "electrode shank":
            r_z = [[-1E199, -600, -550, 1E99],[0, 0, 10, 10]],
    ),
    
    #runtime, cell-specific attributes and output that will be stored
    savelist = [
        'somav',
        'somapos',
        'x',
        'y',
        'z',
        'LFP',
        'CSD',
    ],
    
    #flag for switching on calculation of CSD
    calculateCSD = True,
    
    #time resolution of saved signals
    dt_output = 1. 
))

 
#for each population, define layer- and population-specific connectivity
#parameters
PS.update(dict(
    #number of connections from each presynaptic population onto each
    #layer per postsynaptic population, preserving overall indegree
    k_yXL = dict(
        EX = [[int(0.5*BN.CE), 0],
              [int(0.5*BN.CE), BN.CI]],
        IN = [[0, 0],
              [BN.CE, BN.CI]],
    ),
    
    #set up table of synapse  from each possible 
    J_yX = dict(
        EX = [BN.J_ex*1E-3, BN.J_in*1E-3],
        IN = [BN.J_ex*1E-3, BN.J_in*1E-3],
    ),

    #set up synapse parameters as derived from the network
    synParams = dict(
        EX = dict(
            section = ['apic', 'dend'],
            tau = BN.tauSyn,
            syntype = 'AlphaISyn'
        ),
        IN = dict(
            section = ['dend', 'soma'],
            tau = BN.tauSyn,
            syntype = 'AlphaISyn'            
        ),
    ),
    
    #set up delays, here using fixed delays of network
    synDelayLoc = dict(
        EX = [BN.delay, BN.delay],
        IN = [BN.delay, BN.delay],
    ),
    #no distribution of delays
    synDelayScale = dict(
        EX = [None, None],
        IN = [None, None],
    ),
    
    
))


#putative mappting between population type and cell type specificity,
#but here all presynaptic senders are also postsynaptic targets
PS.update(dict(
    mapping_Yy = list(zip(PS.X, PS.X))
))

################################################################################
# MAIN simulation procedure                                                    #
################################################################################

#tic toc
tic = time()

######### Perform network simulation ###########################################
if properrun:
    #set up the file destination, removing old results by default
    setup_file_dest(PS, clearDestination=True)

if properrun:
    #execute network simulation
    BN.simulate()

#wait for the network simulation to finish, resync MPI threads
COMM.Barrier()


#Create an object representation containing the spiking activity of the network
#simulation output that uses sqlite3. Again, kwargs are derived from the brunel
#network instance.
networkSim = CachedNetwork(
    simtime = BN.simtime,
    dt = BN.dt,
    spike_output_path = BN.spike_output_path,
    label = BN.label,
    ext = 'gdf',
    GIDs = {'EX' : [1, BN.NE], 'IN' : [BN.NE+1, BN.NI]},
    cmap='rainbow_r',
)


if RANK == 0:
    toc = time() - tic
    print('NEST simulation and gdf file processing done in  %.3f seconds' % toc)


####### Set up populations #####################################################

if properrun:
    #iterate over each cell type, and create populationulation object
    for i, Y in enumerate(PS.X):
        #create population:
        pop = Population(
                cellParams = PS.cellParams[Y],
                rand_rot_axis = PS.rand_rot_axis[Y],
                simulationParams = PS.simulationParams,
                populationParams = PS.populationParams[Y],
                y = Y,
                layerBoundaries = PS.layerBoundaries,
                electrodeParams = PS.electrodeParams,
                savelist = PS.savelist,
                savefolder = PS.savefolder,
                calculateCSD = PS.calculateCSD,
                dt_output = PS.dt_output, 
                POPULATIONSEED = SIMULATIONSEED + i,
                X = PS.X,
                networkSim = networkSim,
                k_yXL = PS.k_yXL[Y],
                synParams = PS.synParams[Y],
                synDelayLoc = PS.synDelayLoc[Y],
                synDelayScale = PS.synDelayScale[Y],
                J_yX = PS.J_yX[Y],
            )
    
        #run population simulation and collect the data
        pop.run()
        pop.collect_data()
    
        #object no longer needed
        del pop


####### Postprocess the simulation output ######################################


#reset seed, but output should be deterministic from now on
np.random.seed(SIMULATIONSEED)

if properrun:
    #do some postprocessing on the collected data, i.e., superposition
    #of population LFPs, CSDs etc
    postproc = PostProcess(y = PS.X,
                           dt_output = PS.dt_output,
                           savefolder = PS.savefolder,
                           mapping_Yy = PS.mapping_Yy,
                           savelist = PS.savelist,
                           cells_subfolder = os.path.split(PS.cells_path)[-1],
                           populations_subfolder = os.path.split(PS.populations_path)[-1],
                           figures_subfolder = os.path.split(PS.figures_path)[-1]
                           )
    
    #run through the procedure
    postproc.run()
    
    #create tar-archive with output for plotting, ssh-ing etc.
    postproc.create_tar_archive()
    
    
COMM.Barrier()

#tic toc
print('Execution time: %.3f seconds' %  (time() - tic))



################################################################################
# Create set of plots from simulation output
################################################################################

#import some plotter functions
from example_plotting import *

#turn off interactive plotting
plt.ioff()

if RANK == 0:
    #create network raster plot
    fig = networkSim.raster_plots(xlim=(500, 1000), markersize=2.)
    fig.savefig(os.path.join(PS.figures_path, 'network.pdf'), dpi=300)
    

    #plot cell locations
    fig, ax = plt.subplots(1,1, figsize=(5,8))
    plot_population(ax, PS.populationParams, PS.electrodeParams,
                    PS.layerBoundaries,
                    X=['EX', 'IN'], markers=['^', 'o'], colors=['r', 'b'],
                    layers = ['upper', 'lower'],
                    isometricangle=np.pi/12, aspect='equal')
    fig.savefig(os.path.join(PS.figures_path, 'layers.pdf'), dpi=300)
    

    #plot cell locations
    fig, ax = plt.subplots(1,1, figsize=(5,8))
    plot_population(ax, PS.populationParams, PS.electrodeParams,
                    PS.layerBoundaries,
                    X=['EX', 'IN'], markers=['^', 'o'], colors=['r', 'b'],
                    layers = ['upper', 'lower'],
                    isometricangle=np.pi/12, aspect='equal')
    plot_soma_locations(ax, X=['EX', 'IN'],
                        populations_path=PS.populations_path,
                        markers=['^', 'o'], colors=['r', 'b'],
                        isometricangle=np.pi/12, )
    fig.savefig(os.path.join(PS.figures_path, 'soma_locations.pdf'), dpi=300)
    

    #plot morphologies in their respective locations
    fig, ax = plt.subplots(1,1, figsize=(5,8))
    plot_population(ax, PS.populationParams, PS.electrodeParams,
                    PS.layerBoundaries,
                    X=['EX', 'IN'], markers=['^', 'o'], colors=['r', 'b'],
                    layers = ['upper', 'lower'],
                    aspect='equal')
    plot_morphologies(ax, X=['EX', 'IN'], markers=['^', 'o'], colors=['r', 'b'],
                    isometricangle=np.pi/12,
                    populations_path=PS.populations_path,
                    cellParams=PS.cellParams)
    fig.savefig(os.path.join(PS.figures_path, 'populations.pdf'), dpi=300)
    

    #plot morphologies in their respective locations
    fig, ax = plt.subplots(1,1, figsize=(5,8))
    plot_population(ax, PS.populationParams, PS.electrodeParams,
                    PS.layerBoundaries,
                    X=['EX', 'IN'], markers=['^', 'o'], colors=['r', 'b'],
                    layers = ['upper', 'lower'],
                    aspect='equal')
    plot_individual_morphologies(ax, X=['EX', 'IN'], markers=['^', 'o'],
                                 colors=['r', 'b'],
                                 isometricangle=np.pi/12,
                                 cellParams=PS.cellParams,
                                 populationParams=PS.populationParams)
    fig.savefig(os.path.join(PS.figures_path, 'cell_models.pdf'), dpi=300)
    

    #plot EX morphologies in their respective locations
    fig, ax = plt.subplots(1,1, figsize=(5,8))
    plot_population(ax, PS.populationParams, PS.electrodeParams,
                    PS.layerBoundaries,
                    X=['EX'], markers=['^'], colors=['r'],
                    layers = ['upper', 'lower'],
                    aspect='equal')
    plot_morphologies(ax, X=['EX'], markers=['^'], colors=['r'],
                    isometricangle=np.pi/12,
                    populations_path=PS.populations_path,
                    cellParams=PS.cellParams)
    fig.savefig(os.path.join(PS.figures_path, 'EX_population.pdf'), dpi=300)


    #plot IN morphologies in their respective locations
    fig, ax = plt.subplots(1,1, figsize=(5,8))
    plot_population(ax, PS.populationParams, PS.electrodeParams,
                    PS.layerBoundaries,
                    X=['IN'], markers=['o'], colors=['b'],
                    layers = ['upper', 'lower'],
                    isometricangle=np.pi/12, aspect='equal')
    plot_morphologies(ax, X=['IN'], markers=['o'], colors=['b'],
                    isometricangle=np.pi/12,
                    populations_path=PS.populations_path,
                    cellParams=PS.cellParams)
    fig.savefig(os.path.join(PS.figures_path, 'IN_population.pdf'), dpi=300)

    
    #plot compound LFP and CSD traces
    fig = plt.figure()
    gs = gridspec.GridSpec(2,8)
    
    ax0 = fig.add_subplot(gs[:,:2])
    ax1 = fig.add_subplot(gs[0, 4:])
    ax2 = fig.add_subplot(gs[1, 4:])
    ax1.set_title('CSD')
    ax2.set_title('LFP')

    plot_population(ax0, PS.populationParams, PS.electrodeParams,
                    PS.layerBoundaries,
                    X=['EX', 'IN'], markers=['^', 'o'], colors=['r', 'b'],
                    layers = ['upper', 'lower'],
                    isometricangle=np.pi/12, aspect='equal')
    plot_morphologies(ax0, X=['EX', 'IN'], markers=['^', 'o'],
                      colors=['r', 'b'],
                    isometricangle=np.pi/12,
                    populations_path=PS.populations_path,
                    cellParams=PS.cellParams)

    plot_signal_sum(ax1, z=PS.electrodeParams['z'],
                    fname=os.path.join(PS.savefolder, 'CSDsum.h5'),
                    unit='$\mu$Amm$^{-3}$', T=(500, 1000))
    
    plot_signal_sum(ax2, z=PS.electrodeParams['z'],
                    fname=os.path.join(PS.savefolder, 'LFPsum.h5'),
                    unit='mV', T=(500, 1000))
    
    fig.savefig(os.path.join(PS.figures_path, 'compound_signals.pdf'), dpi=300)


    #plot compound LFP and CSD traces
    fig = plt.figure()
    gs = gridspec.GridSpec(2,8)
    
    ax0 = fig.add_subplot(gs[:,:2])
    ax1 = fig.add_subplot(gs[0, 4:])
    ax2 = fig.add_subplot(gs[1, 4:])
    ax1.set_title('CSD')
    ax2.set_title('LFP')

    plot_population(ax0, PS.populationParams, PS.electrodeParams,
                    PS.layerBoundaries,
                    X=['EX'], markers=['^'], colors=['r'],
                    layers = ['upper', 'lower'],
                    isometricangle=np.pi/12, aspect='equal')
    plot_morphologies(ax0, X=['EX'], markers=['^'], colors=['r'],
                    isometricangle=np.pi/12,
                    populations_path=PS.populations_path,
                    cellParams=PS.cellParams)

    plot_signal_sum(ax1, z=PS.electrodeParams['z'],
                    fname=os.path.join(PS.populations_path,
                                       'EX_population_CSD.h5'),
                    unit='$\mu$Amm$^{-3}$', T=(500, 1000),color='r')
    
    plot_signal_sum(ax2, z=PS.electrodeParams['z'],
                    fname=os.path.join(PS.populations_path,
                                       'EX_population_LFP.h5'),
                    unit='mV', T=(500, 1000), color='r')
    fig.savefig(os.path.join(PS.figures_path, 'population_EX_signals.pdf'),
                dpi=300)


    #plot compound LFP and CSD traces
    fig = plt.figure()
    gs = gridspec.GridSpec(2,8)
    
    ax0 = fig.add_subplot(gs[:,:2])
    ax1 = fig.add_subplot(gs[0, 4:])
    ax2 = fig.add_subplot(gs[1, 4:])
    ax1.set_title('CSD')
    ax2.set_title('LFP')

    plot_population(ax0, PS.populationParams, PS.electrodeParams,
                    PS.layerBoundaries,
                    X=['IN'], markers=['o'], colors=['b'],
                    layers = ['upper', 'lower'],
                    isometricangle=np.pi/12, aspect='equal')
    plot_morphologies(ax0, X=['IN'], markers=['o'], colors=['b'],
                    isometricangle=np.pi/12,
                    populations_path=PS.populations_path,
                    cellParams=PS.cellParams)

    plot_signal_sum(ax1, z=PS.electrodeParams['z'],
                    fname=os.path.join(PS.populations_path,
                                       'IN_population_CSD.h5'),
                    unit='$\mu$Amm$^{-3}$', T=(500, 1000),color='b')
    
    plot_signal_sum(ax2, z=PS.electrodeParams['z'],
                    fname=os.path.join(PS.populations_path,
                                       'IN_population_LFP.h5'),
                    unit='mV', T=(500, 1000), color='b')
    fig.savefig(os.path.join(PS.figures_path, 'population_IN_signals.pdf'),
                dpi=300)


    #correlate global firing rate of network with CSD/LFP across channels
    #compute firing rates
    x, y = networkSim.get_xy((0, BN.simtime))
    bins = np.arange(0, BN.simtime + 1)
        
    xx = np.r_[x['EX'], x['IN']]
        
            
    fig = plt.figure()
    fig.subplots_adjust(hspace=0.5, wspace=0.5)
    gs = gridspec.GridSpec(4,3)
    
    ax0 = fig.add_subplot(gs[0, :2])
    ax1 = fig.add_subplot(gs[1:, :2])
    ax2 = fig.add_subplot(gs[:, 2])

    ax0.hist(xx, bins=bins[500:], color='r')
    ax0.axis('tight')
    ax0.set_title('spike rate (s$^{-1}$)')

    plot_signal_sum(ax1, z=PS.electrodeParams['z'],
                    fname=os.path.join(PS.savefolder, 'LFPsum.h5'),
                    unit='mV', T=(500, 1000))
    ax1.set_title('LFP')

    fname=os.path.join(PS.savefolder, 'LFPsum.h5')
    f = h5py.File(fname, 'r')
    data = f['data'].value
    
    r, t = np.histogram(xx, bins)
    plot_correlation(z_vec=PS.electrodeParams['z'], x0=r, x1=data[:, 1:],
                     ax=ax2, lag=50, title='rate-LFP xcorr')

    fig.savefig(os.path.join(PS.figures_path,
                             'compound_signal_correlations.pdf'), dpi=300)


    #plot morphologies in their respective locations
    fig, ax = plt.subplots(1,1, figsize=(5,8))
    plot_population(ax, PS.populationParams, PS.electrodeParams,
                    PS.layerBoundaries,
                    X=['EX', 'IN'], markers=['^', 'o'], colors=['r', 'b'],
                    layers = ['upper', 'lower'],
                    aspect='equal')
    plot_morphologies(ax, X=['EX', 'IN'], markers=['^', 'o'], colors=['r', 'b'],
                    isometricangle=np.pi/12,
                    populations_path=PS.populations_path,
                    cellParams=PS.cellParams)

    #keep current axes bounds, so we can put it back
    axis = ax.axis()
    
    #some additional plot annotations
    ax.text(-275, -300, 'EX', clip_on=False, va='center', zorder=500)
    ax.add_patch(plt.Rectangle((-290, -340), fc='r', ec='k', alpha=0.5,
        width=80, height=80, clip_on=False, zorder=500))
    ax.arrow(-210, -300, 50, 50, head_width=20, head_length=20, width=10,
             fc='r', lw=1, ec='w', alpha=1, zorder=500)
    ax.arrow(-210, -300, 50, -50, head_width=20, head_length=20, width=10,
             fc='r', lw=1, ec='w', alpha=1, zorder=500)
    
    ax.text(-275, -400, 'IN', clip_on=False, va='center', zorder=500)
    ax.add_patch(plt.Rectangle((-290, -440), fc='b', ec='k', alpha=0.5,
        width=80, height=80, clip_on=False, zorder=500))
    ax.arrow(-210, -400, 50, 0, head_width=20, head_length=20, width=10, fc='b',
             lw=1, ec='w', alpha=1, zorder=500)    
    
    fig.savefig(os.path.join(PS.figures_path, 'populations_vII.pdf'), dpi=300)


    if SIZE == 1:
        plt.show()
    else:
        plt.close('all')

COMM.Barrier()






#!/usr/bin/env python
"""
Function true_lam_csd specification for calculation of true laminar CSD
from the current-distribution on `LFPy.cell.Cell` objects, assuming line
sources for each individual compartment, including the soma.
"""
import numpy as np


def _PrPz(r0, z0, r1, z1, r2, z2, r3, z3):
    """
    Intersection point for infinite lines.
    
    Parameters
    ----------
    r0 : float
    z0 : float
    r1 : float
    z1 : float
    r2 : float
    z2 : float
    r3 : float
    z3 : float

    Returns
    ----------
    Pr : float    
    Pz : float
    hit : bool

    """
    Pr = ((r0*z1 - z0*r1)*(r2 - r3) - (r0 - r1)*(r2*z3 - r3*z2)) / \
                        ((r0 - r1)*(z2 - z3) - (z0 - z1)*(r2-r3))
    Pz = ((r0*z1 - z0*r1)*(z2 - z3) - (z0 - z1)*(r2*z3 - r3*z2)) / \
                        ((r0 - r1)*(z2 - z3) - (z0 - z1)*(r2-r3))
    
    if Pr >= r0 and Pr <= r1 and Pz >= z0 and Pz <= z1:
        hit = True
    elif Pr <= r0 and Pr >= r1 and Pz >= z0 and Pz <= z1:
        hit = True
    elif Pr >= r0 and Pr <= r1 and Pz <= z0 and Pz >= z1:
        hit = True
    elif Pr <= r0 and Pr >= r1 and Pz <= z0 and Pz >= z1:
        hit = True
    else:
        hit = False
        
    return [Pr, Pz, hit]


def true_lam_csd(cell, dr=100, z=None):
    """
    Return CSD from membrane currents as function along the coordinates
    of the electrode along z-axis. 


    Parameters
    ----------
    cell : `LFPy.cell.Cell` or `LFPy.cell.TemplateCell` object.
        Cell.
    dr : float
        Radius of the cylindrical CSD volume.
    z : numpy.ndarray
        Z-coordinates of electrode.


    Returns
    ----------
    CSD : numpy.ndarray
        Current-source density (in pA * mum^-3).
    
    """
    if type(z) != type(np.ndarray(shape=0)):
        raise ValueError('type(z) should be a numpy.ndarray')

    dz = abs(z[1] - z[0])
    CSD = np.zeros((z.size, cell.tvec.size,))
    r_end = np.sqrt(cell.xend**2 + cell.yend**2)
    r_start = np.sqrt(cell.xstart**2 + cell.ystart**2)
    volume = dz * np.pi * dr**2

    for i in range(len(z)):
        aa0 = cell.zstart < z[i] + dz/2
        aa1 = cell.zend < z[i] + dz/2
        bb0 = cell.zstart >= z[i] - dz/2
        bb1 = cell.zend >= z[i] - dz/2
        cc0 = r_start < dr
        cc1 = r_end < dr
        ii = aa0 & bb0 & cc0 # startpoint inside volume
        jj = aa1 & bb1 & cc1 # endpoint inside volume

        for j in range(cell.zstart.size):
            # Calc fraction of source being inside control volume from 0-1
            # both start and endpoint in volume
            if ii[j] and jj[j]:
                CSD[i,] = CSD[i, ] + cell.imem[j, ] / volume
            # Startpoint in volume:
            elif ii[j] and not jj[j]: 
                z0 = cell.zstart[j]
                r0 = r_start[j]
                
                z1 = cell.zend[j]
                r1 = r_end[j]

                if r0 == r1:
                    # Segment is parallel with z-axis
                    frac = -(z0 - z[i]-dz/2) / (z1 - z0)
                else:
                    # Not parallel with z-axis                    
                    L2 = (r1 - r0)**2 + (z1 - z0)**2
    
                    z2 = [z[i]-dz/2, z[i]+dz/2, z[i]-dz/2]
                    r2 = [0, 0, dr]
                    
                    z3 = [z[i]-dz/2, z[i]+dz/2, z[i]+dz/2]
                    r3 = [dr, dr, dr]
    
                    P = []
                    for k in range(3):
                        P.append(_PrPz(r0, z0, r1, z1, r2[k], z2[k], r3[k], z3[k]))
                        if P[k][2]:
                            vL2 = (P[k][0] - r0)**2 + (P[k][1] -z0)**2
                            frac = np.sqrt(vL2 / L2)
                CSD[i,] = CSD[i, ] + frac * cell.imem[j, ] / volume
            # Endpoint in volume:
            elif jj[j] and not ii[j]: 
                z0 = cell.zstart[j]
                r0 = r_start[j]
                
                z1 = cell.zend[j]
                r1 = r_end[j]
                
                if r0 == r1:
                    # Segment is parallel with z-axis
                    frac = (z1 - z[i]+dz/2) / (z1 - z0)
                else:
                    # Not parallel with z-axis
                    L2 = (r1 - r0)**2 + (z1 - z0)**2
    
                    z2 = [z[i]-dz/2, z[i]+dz/2, z[i]-dz/2]
                    r2 = [0, 0, dr]
                    
                    z3 = [z[i]-dz/2, z[i]+dz/2, z[i]+dz/2]
                    r3 = [dr, dr, dr]
    
                    P = []
                    for k in range(3):
                        P.append(_PrPz(r0, z0, r1, z1, r2[k], z2[k], r3[k], z3[k]))
                        if P[k][2]:
                            vL2 = (r1 - P[k][0])**2 + (z1 - P[k][1])**2
                            frac = np.sqrt(vL2 / L2)
                CSD[i,] = CSD[i, ] + frac * cell.imem[j, ] / volume
            else:
                pass

    return CSD


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/env/bin python
#
# brunel_alpha_nest.py
#
# This file is part of NEST.
#
# Copyright (C) 2004 The NEST Initiative
#
# NEST is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# NEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with NEST.  If not, see <http://www.gnu.org/licenses/>.
#
# Modified as an example with hybridLFPy (https:/github.com/INM-6/hybridLFPy)
# for calculating local field potentials from spiking point-neuron networks

'''
Random balanced network (alpha synapses) connected with NEST
------------------------------------------------------------

This script simulates an excitatory and an inhibitory population on
the basis of the network used in

Brunel N, Dynamics of Sparsely Connected Networks of Excitatory and
Inhibitory Spiking Neurons, Journal of Computational Neuroscience 8,
183-208 (2000).

In contrast to brunel-alpha-numpy.py, this variant uses NEST's builtin
connection routines to draw the random connections instead of NumPy.

When connecting the network customary synapse models are used, which
allow for querying the number of created synapses. Using spike
detectors the average firing rates of the neurons in the populations
are established. The building as well as the simulation time of the
network are recorded.
'''

'''
Importing all necessary modules for simulation, analysis and plotting.
'''
from scipy.optimize import fsolve

import nest
import nest.raster_plot

import os
import time
from numpy import exp, random

'''
Definition of functions used in this example. First, define the
Lambert W function implemented in SLI. The second function
computes the maximum of the postsynaptic potential for a synaptic
input current of unit amplitude (1 pA) using the Lambert W
function. Thus function will later be used to calibrate the synaptic
weights.
'''

def LambertWm1(x):
    nest.sli_push(x); nest.sli_run('LambertWm1'); y=nest.sli_pop()
    return y

def ComputePSPnorm(tauMem, CMem, tauSyn):
  a = (tauMem / tauSyn)
  b = (1.0 / tauSyn - 1.0 / tauMem)

  # time of maximum
  t_max = 1.0/b * ( -LambertWm1(-exp(-1.0/a)/a) - 1.0/a )

  # maximum of PSP for current of unit amplitude
  return exp(1.0)/(tauSyn*CMem*b) * ((exp(-t_max/tauMem) - exp(-t_max/tauSyn)) / b - t_max*exp(-t_max/tauSyn))

'''
Assigning the current time to a variable in order to determine the
build time of the network.
'''

startbuild = time.time()

'''
Assigning the simulation parameters to variables.
'''

dt      = 0.1    # the resolution in ms
simtime = 1000.0 # Simulation time in ms
delay   = 1.5    # synaptic delay in ms

'''
Definition of the parameters crucial for asynchronous irregular firing
of the neurons.
'''

g       = 6.0  # ratio inhibitory weight/excitatory weight
eta     = 2.0  # external rate relative to threshold rate
epsilon = 0.1  # connection probability

'''
Definition of the number of neurons in the network and the number of
neuron recorded from.

Note: We record here all spike events.
'''

order     = 100
NE        = 4*order # number of excitatory neurons
NI        = 1*order # number of inhibitory neurons
N_neurons = NE+NI   # number of neurons in total

'''
flag for using Poisson or equivalent DC current external drive
'''

Poisson = False

'''
Definition of connectivity parameter
'''

CE    = int(epsilon*NE) # number of excitatory synapses per neuron
CI    = int(epsilon*NI) # number of inhibitory synapses per neuron
C_tot = int(CI+CE)      # total number of synapses per neuron

'''
Initialization of the parameters of the integrate and fire neuron and
the synapses. The parameter of the neuron are stored in a dictionary.
The synaptic currents are normalized such that the amplitude of the
PSP is J.
'''

tauSyn = 0.5        # synaptic time constant in ms
tauMem = 20.0       # time constant of membrane potential in ms
CMem = 250.0        # capacitance of membrane in in pF
theta  = 20.0       # membrane threshold potential in mV
J      = 1.0        # postsynaptic amplitude in mV
J_unit = ComputePSPnorm(tauMem, CMem, tauSyn)
J_ex   = J / J_unit # amplitude of excitatory postsynaptic current
J_in   = -g*J_ex    # amplitude of inhibitory postsynaptic current

'''
Definition of threshold rate, which is the external rate needed to fix
the membrane potential around its threshold, the external firing rate
and the rate of the poisson generator which is multiplied by the
in-degree CE and converted to Hz by multiplication by 1000.
'''

nu_th  = (theta * CMem) / (J_ex*CE*exp(1)*tauMem*tauSyn)
nu_ex  = eta*nu_th
p_rate = 1000.0*nu_ex*CE

'''
Adjust parameters accordingly if Poisson background input is used
'''

if Poisson:
    #use zero dc input amplitude to the neurons
    dc_amplitude = 0.
else:
    #compute equivalent DC current amplitude instead of Poisson input assuming
    #alpha current synapse
    factor = 0.4 #####CHECK
    dc_amplitude = p_rate*J_ex / 1000 / tauSyn * factor

'''
Parameters for "iaf_psc_alpha" cell model
'''

neuron_params= {"C_m":        CMem,
                "tau_m":      tauMem,
                "tau_syn_ex": tauSyn,
                "tau_syn_in": tauSyn,
                "t_ref":      2.0,
                "E_L":        0.0,
                "V_reset":    0.0,
                "V_th":       20.0,
                "V_m":        0.0,
                "V_th":       theta,
                "I_e":        dc_amplitude,}
'''
Destination of file output
'''

spike_output_path = os.path.join('simulation_output_example_brunel',
                                 'spiking_output_path')

'''
File prefix for spike detectors
'''

label = 'brunel-py'

'''
Main simulation procedure function
'''

def simulate():
    '''instantiate and execute network simulation'''
    #separate model execution from parameters for safe import from other files
    nest.ResetKernel()

    '''
    Configuration of the simulation kernel by the previously defined time
    resolution used in the simulation. Setting "print_time" to True prints
    the already processed simulation time as well as its percentage of the
    total simulation time.
    '''

    nest.SetKernelStatus({"resolution": dt, "print_time": True,
                          "overwrite_files": True})
    
    print("Building network")
    
    
    '''
    Configuration of the model `iaf_psc_alpha` and `poisson_generator`
    using SetDefaults(). This function expects the model to be the
    inserted as a string and the parameter to be specified in a
    dictionary. All instances of theses models created after this point
    will have the properties specified in the dictionary by default.
    '''
    
    nest.SetDefaults("iaf_psc_alpha", neuron_params)
    nest.SetDefaults("poisson_generator",{"rate": p_rate})
    
    '''
    Creation of the nodes using `Create`. We store the returned handles in
    variables for later reference. Here the excitatory and inhibitory, as
    well as the poisson generator and two spike detectors. The spike
    detectors will later be used to record excitatory and inhibitory
    spikes.
    '''
    
    nodes_ex = nest.Create("iaf_psc_alpha",NE)
    nodes_in = nest.Create("iaf_psc_alpha",NI)
    noise    = nest.Create("poisson_generator")
    espikes  = nest.Create("spike_detector")
    ispikes  = nest.Create("spike_detector")
    
    print("first exc node: {}".format(nodes_ex[0]))
    print("first inh node: {}".format(nodes_in[0]))
    
    '''
    distribute membrane potentials
    '''
    nest.SetStatus(nodes_ex, "V_m",
                   random.rand(len(nodes_ex))*neuron_params["V_th"])
    nest.SetStatus(nodes_in, "V_m",
                   random.rand(len(nodes_in))*neuron_params["V_th"])
    
    '''
    Configuration of the spike detectors recording excitatory and
    inhibitory spikes using `SetStatus`, which expects a list of node
    handles and a list of parameter dictionaries. Setting the variable
    "to_file" to True ensures that the spikes will be recorded in a .gdf
    file starting with the string assigned to label. Setting "withtime"
    and "withgid" to True ensures that each spike is saved to file by
    stating the gid of the spiking neuron and the spike time in one line.
    '''
    
    nest.SetStatus(espikes,[{
                       "label": os.path.join(spike_output_path, label + "-EX"),
                       "withtime": True,
                       "withgid": True,
                       "to_file": True,
                       }])
    
    nest.SetStatus(ispikes,[{
                       "label": os.path.join(spike_output_path, label + "-IN"),
                       "withtime": True,
                       "withgid": True,
                       "to_file": True,}])
    
    print("Connecting devices")
    
    '''
    Definition of a synapse using `CopyModel`, which expects the model
    name of a pre-defined synapse, the name of the customary synapse and
    an optional parameter dictionary. The parameters defined in the
    dictionary will be the default parameter for the customary
    synapse. Here we define one synapse for the excitatory and one for the
    inhibitory connections giving the previously defined weights and equal
    delays.
    '''
    
    nest.CopyModel("static_synapse","excitatory",{"weight":J_ex, "delay":delay})
    nest.CopyModel("static_synapse","inhibitory",{"weight":J_in, "delay":delay})
    
    '''
    Connecting the previously defined poisson generator to the excitatory
    and inhibitory neurons using the excitatory synapse. Since the poisson
    generator is connected to all neurons in the population the default
    rule ('all_to_all') of Connect() is used. The synaptic properties are
    inserted via syn_spec which expects a dictionary when defining
    multiple variables or a string when simply using a pre-defined
    synapse.
    '''
    
    if Poisson:
      nest.Connect(noise,nodes_ex, 'all_to_all', "excitatory")
      nest.Connect(noise,nodes_in,'all_to_all', "excitatory")
    
    '''
    Connecting the first N_neurons nodes of the excitatory and inhibitory
    population to the associated spike detectors using excitatory
    synapses. Here the same shortcut for the specification of the synapse
    as defined above is used.
    '''
    
    nest.Connect(nodes_ex,espikes, 'all_to_all', "excitatory")
    nest.Connect(nodes_in,ispikes, 'all_to_all', "excitatory")
    
    print("Connecting network")
    
    print("Excitatory connections")
    
    '''
    Connecting the excitatory population to all neurons using the
    pre-defined excitatory synapse. Beforehand, the connection parameter
    are defined in a dictionary. Here we use the connection rule
    'fixed_indegree', which requires the definition of the indegree. Since
    the synapse specification is reduced to assigning the pre-defined
    excitatory synapse it suffices to insert a string.
    '''
    
    conn_params_ex = {'rule': 'fixed_indegree', 'indegree': CE}
    nest.Connect(nodes_ex, nodes_ex+nodes_in, conn_params_ex, "excitatory")
    
    print("Inhibitory connections")
    
    '''
    Connecting the inhibitory population to all neurons using the
    pre-defined inhibitory synapse. The connection parameter as well as
    the synapse paramtere are defined analogously to the connection from
    the excitatory population defined above.
    '''
    
    conn_params_in = {'rule': 'fixed_indegree', 'indegree': CI}
    nest.Connect(nodes_in, nodes_ex+nodes_in, conn_params_in, "inhibitory")
    
    
    '''
    Storage of the time point after the buildup of the network in a
    variable.
    '''
    
    endbuild=time.time()
    
    '''
    Simulation of the network.
    '''
    
    print("Simulating")
    
    nest.Simulate(simtime)
    
    '''
    Storage of the time point after the simulation of the network in a
    variable.
    '''
    
    endsimulate= time.time()
    
    '''
    Reading out the total number of spikes received from the spike
    detector connected to the excitatory population and the inhibitory
    population.
    '''
    
    events_ex = nest.GetStatus(espikes,"n_events")[0]
    events_in = nest.GetStatus(ispikes,"n_events")[0]
    
    '''
    Calculation of the average firing rate of the excitatory and the
    inhibitory neurons by dividing the total number of recorded spikes by
    the number of neurons recorded from and the simulation time. The
    multiplication by 1000.0 converts the unit 1/ms to 1/s=Hz.
    '''
    
    rate_ex   = events_ex/simtime*1000.0/N_neurons
    rate_in   = events_in/simtime*1000.0/N_neurons
    
    '''
    Reading out the number of connections established using the excitatory
    and inhibitory synapse model. The numbers are summed up resulting in
    the total number of synapses.
    '''
    
    num_synapses = nest.GetDefaults("excitatory")["num_connections"]+\
    nest.GetDefaults("inhibitory")["num_connections"]
    
    '''
    Establishing the time it took to build and simulate the network by
    taking the difference of the pre-defined time variables.
    '''
    
    build_time = endbuild-startbuild
    sim_time   = endsimulate-endbuild
    
    '''
    Printing the network properties, firing rates and building times.
    '''
    
    print("Brunel network simulation (Python)")
    print("Number of neurons : {0}".format(N_neurons))
    print("Number of synapses: {0}".format(num_synapses))
    print("       Exitatory  : {0}".format(int(CE * N_neurons) + N_neurons))
    print("       Inhibitory : {0}".format(int(CI * N_neurons)))
    print("Excitatory rate   : %.2f Hz" % rate_ex)
    print("Inhibitory rate   : %.2f Hz" % rate_in)
    print("Building time     : %.2f s" % build_time)
    print("Simulation time   : %.2f s" % sim_time)
    
    '''
    Plot a raster of the excitatory neurons and a histogram.
    '''
    
    if False:
        nest.raster_plot.from_device(espikes, hist=True)
        nest.raster_plot.from_device(ispikes, hist=True)
        nest.raster_plot.show()

if __name__ == '__main__':
    simulate()



#!/usr/bin/env python
"""
Class methods defining multicompartment neuron populations in the hybrid scheme
"""
import os
import glob
import numpy as np
import h5py
import matplotlib.pyplot as plt
from mpi4py import MPI
from .gdf import GDF
from . import csd
from . import helpers
import LFPy
import neuron
from time import time


################# Initialization of MPI stuff ##################################
COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()


############ class objects #####################################################
class PopulationSuper(object):
    """
    Main population class object, let one set up simulations, execute, and
    compile the results. This class is suitable for subclassing for
    custom cell simulation procedures, inherit things like gathering of
    data written to disk.

    Note that `PopulationSuper.cellsim` do not have any stimuli,
    its main purpose is to gather common methods for inherited Population
    objects.


    Parameters
    ----------
    cellParams : dict
        Parameters for class `LFPy.Cell`
    rand_rot_axis : list
        Axis of which to randomly rotate morphs.
    simulationParams : dict
        Additional args for `LFPy.Cell.simulate()`.
    populationParams : dict
        Constraints for population and cell number.
    y : str
        Population identifier string.
    layerBoundaries : list of lists
        Each element is a list setting upper and lower layer boundary (floats)
    electrodeParams : dict
        parameters for class `LFPy.RecExtElectrode`
    savelist : list
        `LFPy.Cell` arguments to save for each single-cell simulation.
    savefolder : str
        path to where simulation results are stored.
    calculateCSD : bool
        Exctract laminar CSD from transmembrane currents
    dt_output : float
        Time resolution of output, e.g., LFP, CSD etc.
    recordSingleContribFrac : float
        fraction  in [0, 1] of individual neurons in population which output
        will be stored
    POPULATIONSEED : int/float
        Random seed for population, for positions etc.
    verbose : bool
        Verbosity flag.
    output_file : str
        formattable string for population output, e.g., '{}_population_{}'


    Returns
    -------
    hybridLFPy.population.PopulationSuper object


    See also
    --------
    Population, LFPy.Cell, LFPy.RecExtElectrode

    """

    def __init__(self,
                 cellParams={
                    'morphology': 'morphologies/ex.hoc',
                    'Ra': 150,
                    'cm': 1.0,
                    'e_pas': 0.0,
                    'lambda_f': 100,
                    'nsegs_method': 'lambda_f',
                    'rm': 20000.0,
                    'timeres_NEURON': 0.1,
                    'timeres_python': 0.1,
                    'tstartms': 0,
                    'tstopms': 1000.0,
                    'v_init': 0.0,
                    'verbose': False},
                 rand_rot_axis=[],
                 simulationParams={},
                 populationParams={
                    'min_cell_interdist': 1.0,
                    'number': 400,
                    'radius': 100,
                    'z_max': -350,
                    'z_min': -450},
                 y = 'EX',
                 layerBoundaries = [[0.0, -300], [-300, -500]],
                 electrodeParams={
                    'N': [[1, 0, 0], [1, 0, 0], [1, 0, 0],
                          [1, 0, 0], [1, 0, 0], [1, 0, 0]],
                    'method': 'som_as_point',
                    'n': 20,
                    'r': 5,
                    'r_z': [[-1e+199, -600, -550, 1e+99], [0, 0, 10, 10]],
                    'seedvalue': None,
                    'sigma': 0.3,
                    'x': [0, 0, 0, 0, 0, 0],
                    'y': [0, 0, 0, 0, 0, 0],
                    'z': [-0.0, -100.0, -200.0, -300.0, -400.0, -500.0]},
                 savelist=['somapos', 'x', 'y', 'z', 'LFP', 'CSD'],
                 savefolder='simulation_output_example_brunel',
                 calculateCSD=True,
                 dt_output=1.,
                 recordSingleContribFrac=0,
                 POPULATIONSEED=123456,
                 verbose=False,
                 output_file='{}_population_{}'
                 ):

        """
        Main population class object, let one set up simulations, execute, and
        compile the results. This class is suitable for subclassing for
        custom cell simulation procedures, inherit things like gathering of
        data written to disk.

        Note that `PopulationSuper.cellsim` do not have any stimuli,
        its main purpose is to gather common methods for inherited Population
        objects.


        Parameters
        ----------
        cellParams : dict
            Parameters for class `LFPy.Cell`
        rand_rot_axis : list
            Axis of which to randomly rotate morphs.
        simulationParams : dict
            Additional args for `LFPy.Cell.simulate()`.
        populationParams : dict
            Constraints for population and cell number.
        y : str
            Population identifier string.
        layerBoundaries : list of lists
            Each element is a list setting upper and lower layer boundary as
            floats
        electrodeParams : dict
            parameters for class `LFPy.RecExtElectrode`
        savelist : list
            `LFPy.Cell` arguments to save for each single-cell simulation.
        savefolder : str
            path to where simulation results are stored.
        calculateCSD : bool
            Exctract laminar CSD from transmembrane currents
        dt_output : float
            Time resolution of output, e.g., LFP, CSD etc.
        recordSingleContribFrac : float
            fraction  in [0, 1] of individual neurons in population which output
            will be stored
        POPULATIONSEED : int/float
            Random seed for population, for positions etc.
        verbose : bool
            Verbosity flag.
        output_file : str
            formattable string for population output, e.g., '{}_population_{}'


        Returns
        -------
        hybridLFPy.population.PopulationSuper object


        See also
        --------
        Population, LFPy.Cell, LFPy.RecExtElectrode

        """

        self.cellParams = cellParams
        self.dt = self.cellParams['timeres_python']
        self.rand_rot_axis = rand_rot_axis
        self.simulationParams = simulationParams
        self.populationParams = populationParams
        self.POPULATION_SIZE = populationParams['number']
        self.y = y
        self.layerBoundaries = np.array(layerBoundaries)
        self.electrodeParams = electrodeParams
        self.savelist = savelist
        self.savefolder = savefolder
        self.calculateCSD = calculateCSD
        self.dt_output = dt_output
        self.recordSingleContribFrac = recordSingleContribFrac
        self.output_file = output_file
        
        #check that decimate fraction is actually a whole number
        try:
            assert int(self.dt_output / self.dt) == self.dt_output / self.dt
        except AssertionError as ae:
            raise ae, 'dt_output not an integer multiple of dt'
        
        self.decimatefrac = int(self.dt_output / self.dt)
        self.POPULATIONSEED = POPULATIONSEED
        self.verbose = verbose
        
        #put revision info in savefolder
        if self.savefolder is not None:
            os.system('git rev-parse HEAD -> %s/populationRevision.txt' % \
                    self.savefolder)

        #set the random seed for reproducible populations, synapse locations,
        #presynaptic spiketrains
        np.random.seed(self.POPULATIONSEED)

        #using these colors and alphas:
        self.colors = []
        for i in range(self.POPULATION_SIZE):
            i *= 256.
            if self.POPULATION_SIZE > 1:
                i /= self.POPULATION_SIZE - 1.
            else:
                i /= self.POPULATION_SIZE

            try:
                self.colors.append(plt.cm.rainbow(int(i)))
            except:
                self.colors.append(plt.cm.gist_rainbow(int(i)))


        self.alphas = np.ones(self.POPULATION_SIZE)


        self.pop_soma_pos = self.set_pop_soma_pos()
        self.rotations = self.set_rotations()

        self._set_up_savefolder()

        self.CELLINDICES = np.arange(self.POPULATION_SIZE)
        self.RANK_CELLINDICES = self.CELLINDICES[self.CELLINDICES % SIZE
                                                 == RANK]

        #container for single-cell output generated on this RANK
        self.output = dict((i, {}) for i in self.RANK_CELLINDICES)


    def _set_up_savefolder(self):
        """
        Create catalogs for different file output to clean up savefolder.

        Non-public method


        Parameters
        ----------
        None


        Returns
        -------
        None

        """
        if self.savefolder == None:
            return

        self.cells_path = os.path.join(self.savefolder, 'cells')
        if RANK == 0:
            if not os.path.isdir(self.cells_path):
                os.mkdir(self.cells_path)

        self.figures_path = os.path.join(self.savefolder, 'figures')
        if RANK == 0:
            if not os.path.isdir(self.figures_path):
                os.mkdir(self.figures_path)

        self.populations_path = os.path.join(self.savefolder, 'populations')
        if RANK == 0:
            if not os.path.isdir(self.populations_path):
                os.mkdir(self.populations_path)

        COMM.Barrier()


    def run(self):
        """
        Distribute individual cell simulations across ranks.

        This method takes no keyword arguments.


        Parameters
        ----------
        None


        Returns
        -------
        None

        """
        for cellindex in self.RANK_CELLINDICES:
            self.cellsim(cellindex)

        COMM.Barrier()


    def cellsim(self, cellindex, return_just_cell=False):
        """
        Single-cell `LFPy.Cell` simulation without any stimulus, mostly for
        reference, as no stimulus is added


        Parameters
        ----------
        cellindex : int
            cell index between 0 and POPULATION_SIZE-1.
        return_just_cell : bool
            If True, return only the LFPy.Cell object
            if False, run full simulation, return None.


        Returns
        -------
        None
            if `return_just_cell is False
        cell : `LFPy.Cell` instance
            if `return_just_cell` is True


        See also
        --------
        LFPy.Cell, LFPy.Synapse, LFPy.RecExtElectrode

        """
        tic = time()
        
        electrode = LFPy.RecExtElectrode(**self.electrodeParams)

        cellParams = self.cellParams.copy()
        cell = LFPy.Cell(**cellParams)
        cell.set_pos(**self.pop_soma_pos[cellindex])
        cell.set_rotation(**self.rotations[cellindex])

        if return_just_cell:
            return cell
        else:
            if self.calculateCSD:
                cell.tvec = np.arange(cell.totnsegs)
                cell.imem = np.eye(cell.totnsegs)
                csdcoeff = csd.true_lam_csd(cell,
                                self.populationParams['radius'], electrode.z)
                csdcoeff *= 1E6 #nA mum^-3 -> muA mm^-3 conversion
                del cell.tvec, cell.imem
                cell.simulate(electrode, dotprodcoeffs=[csdcoeff],
                              **self.simulationParams)
                cell.CSD = helpers.decimate(cell.dotprodresults[0],
                                            q=self.decimatefrac)
            else:
                cell.simulate(electrode,
                              **self.simulationParams)

            cell.LFP = helpers.decimate(electrode.LFP,
                                        q=self.decimatefrac)


            cell.x = electrode.x
            cell.y = electrode.y
            cell.z = electrode.z
            cell.electrodecoeff = electrode.electrodecoeff


            #put all necessary cell output in output dict
            for attrbt in self.savelist:
                attr = getattr(cell, attrbt)
                if type(attr) == np.ndarray:
                    self.output[cellindex][attrbt] = attr.astype('float32')
                else:
                    try:
                        self.output[cellindex][attrbt] = attr
                    except:
                        self.output[cellindex][attrbt] = str(attr)
                self.output[cellindex]['srate'] = 1E3 / self.dt_output

            print('cell %s population %s in %.2f s' % (cellindex, self.y,
                                                              time()-tic))


    def set_pop_soma_pos(self):
        """
        Set `pop_soma_pos` using draw_rand_pos().

        This method takes no keyword arguments.


        Parameters
        ----------
        None


        Returns
        -------
        numpy.ndarray
            (x,y,z) coordinates of each neuron in the population


        See also
        --------
        PopulationSuper.draw_rand_pos

        """
        tic = time()
        if RANK == 0:
            pop_soma_pos = self.draw_rand_pos(
                min_r = self.electrodeParams['r_z'],
                **self.populationParams)
        else:
            pop_soma_pos = None

        if RANK == 0:
            print('found cell positions in %.2f s' % (time()-tic))

        return COMM.bcast(pop_soma_pos, root=0)


    def set_rotations(self):

        """

        Append random z-axis rotations for each cell in population.

        This method takes no keyword arguments


        Parameters
        ----------
        None


        Returns
        -------
        numpyp.ndarray
            Rotation angle around axis `Population.rand_rot_axis` of each neuron
            in the population


        """
        tic = time()
        if RANK == 0:
            rotations = []
            for i in range(self.POPULATION_SIZE):
                defaultrot = {}
                for axis in self.rand_rot_axis:
                    defaultrot.update({axis : np.random.rand() * 2 * np.pi})
                rotations.append(defaultrot)
        else:
            rotations = None

        if RANK == 0:
            print('found cell rotations in %.2f s' % (time()-tic))

        return COMM.bcast(rotations, root=0)


    def calc_min_cell_interdist(self, x, y, z):
        """
        Calculate cell interdistance from input coordinates.


        Parameters
        ----------
        x, y, z : numpy.ndarray
            xyz-coordinates of each cell-body.


        Returns
        -------
        min_cell_interdist : np.nparray
            For each cell-body center, the distance to nearest neighboring cell

        """
        min_cell_interdist = np.zeros(self.POPULATION_SIZE)

        for i in range(self.POPULATION_SIZE):
            cell_interdist = np.sqrt((x[i] - x)**2
                    + (y[i] - y)**2
                    + (z[i] - z)**2)
            cell_interdist[i] = np.inf
            min_cell_interdist[i] = cell_interdist.min()

        return min_cell_interdist


    def draw_rand_pos(self, radius, z_min, z_max,
                      min_r=np.array([0]), min_cell_interdist=10., **args):
        """
        Draw some random location within radius, z_min, z_max,
        and constrained by min_r and the minimum cell interdistance.
        Returned argument is a list of dicts [{'xpos', 'ypos', 'zpos'},].


        Parameters
        ----------
        radius : float
            Radius of population.
        z_min : float
            Lower z-boundary of population.
        z_max : float
            Upper z-boundary of population.
        min_r : numpy.ndarray
            Minimum distance to center axis as function of z.
        min_cell_interdist : float
            Minimum cell to cell interdistance.
        **args : keyword arguments
            Additional inputs that is being ignored.


        Returns
        -------
        soma_pos : list
            List of dicts of len population size
            where dict have keys xpos, ypos, zpos specifying
            xyz-coordinates of cell at list entry `i`.


        See also
        --------
        PopulationSuper.calc_min_cell_interdist

        """
        x = (np.random.rand(self.POPULATION_SIZE)-0.5)*radius*2
        y = (np.random.rand(self.POPULATION_SIZE)-0.5)*radius*2
        z = np.random.rand(self.POPULATION_SIZE)*(z_max - z_min) + z_min
        min_r_z = {}
        min_r = np.array(min_r)
        if min_r.size > 0:
            if type(min_r) == type(np.array([])):
                j = 0
                for j in range(min_r.shape[0]):
                    min_r_z[j] = np.interp(z, min_r[0,], min_r[1,])
                    if j > 0:
                        [w] = np.where(min_r_z[j] < min_r_z[j-1])
                        min_r_z[j][w] = min_r_z[j-1][w]
                minrz = min_r_z[j]
        else:
            minrz = np.interp(z, min_r[0], min_r[1])

        R_z = np.sqrt(x**2 + y**2)

        #want to make sure that no somas are in the same place.
        cell_interdist = self.calc_min_cell_interdist(x, y, z)

        [u] = np.where(np.logical_or((R_z < minrz) != (R_z > radius),
            cell_interdist < min_cell_interdist))

        while len(u) > 0:
            for i in range(len(u)):
                x[u[i]] = (np.random.rand()-0.5)*radius*2
                y[u[i]] = (np.random.rand()-0.5)*radius*2
                z[u[i]] = np.random.rand()*(z_max - z_min) + z_min
                if type(min_r) == type(()):
                    for j in range(np.shape(min_r)[0]):
                        min_r_z[j][u[i]] = \
                                np.interp(z[u[i]], min_r[0,], min_r[1,])
                        if j > 0:
                            [w] = np.where(min_r_z[j] < min_r_z[j-1])
                            min_r_z[j][w] = min_r_z[j-1][w]
                        minrz = min_r_z[j]
                else:
                    minrz[u[i]] = np.interp(z[u[i]], min_r[0,], min_r[1,])
            R_z = np.sqrt(x**2 + y**2)

            #want to make sure that no somas are in the same place.
            cell_interdist = self.calc_min_cell_interdist(x, y, z)

            [u] = np.where(np.logical_or((R_z < minrz) != (R_z > radius),
                cell_interdist < min_cell_interdist))

        
        soma_pos = []
        for i in range(self.POPULATION_SIZE):
            soma_pos.append({'xpos' : x[i], 'ypos' : y[i], 'zpos' : z[i]})

        return soma_pos


    def calc_signal_sum(self, measure='LFP'):
        """
        Superimpose each cell's contribution to the compound population signal,
        i.e., the population CSD or LFP


        Parameters
        ----------
        measure : str
            {'LFP', 'CSD'}: Either 'LFP' or 'CSD'.


        Returns
        -------
        numpy.ndarray
            The populations-specific compound signal.

        """
        #compute the total LFP of cells on this RANK
        if self.RANK_CELLINDICES.size > 0:
            for i, cellindex in enumerate(self.RANK_CELLINDICES):
                if i == 0:
                    data = self.output[cellindex][measure]
                else:
                    data += self.output[cellindex][measure]
        else:
            data = np.zeros((len(self.electrodeParams['x']),
                             self.cellParams['tstopms']/self.dt_output + 1),
                dtype=np.float32)

        #container for full LFP on RANK 0
        if RANK == 0:
            DATA = np.zeros_like(data, dtype=np.float32)
        else:
            DATA = None

        #sum to RANK 0 using automatic type discovery with MPI
        COMM.Reduce(data, DATA, op=MPI.SUM, root=0)

        return DATA


    def collectSingleContribs(self, measure='LFP'):
        """
        Collect single cell data and save them to HDF5 file.
        The function will also return signals generated by all cells


        Parameters
        ----------
        measure : str
            {'LFP', 'CSD'}: Either 'LFP' or 'CSD'.


        Returns
        -------
        numpy.ndarray
            output of all neurons in population, axis 0 correspond to neuron ind

        """
        try:
            assert(self.recordSingleContribFrac <= 1 and
                   self.recordSingleContribFrac >= 0)
        except AssertionError as ae:
            raise ae, 'recordSingleContribFrac {} not in [0, 1]'.format(
                                                self.recordSingleContribFrac)

        if not self.recordSingleContribFrac:
            return
        else:
            #reconstruct RANK_CELLINDICES of all RANKs for controlling
            #communication
            if self.recordSingleContribFrac == 1.:
                SAMPLESIZE = self.POPULATION_SIZE
                RANK_CELLINDICES = []
                for i in range(SIZE):
                    RANK_CELLINDICES += [self.CELLINDICES[
                        self.CELLINDICES % SIZE == i]]
            else:
                SAMPLESIZE = int(self.recordSingleContribFrac *
                                 self.POPULATION_SIZE)
                RANK_CELLINDICES = []
                for i in range(SIZE):
                    ids = self.CELLINDICES[self.CELLINDICES % SIZE == i]
                    RANK_CELLINDICES += [ids[ids < SAMPLESIZE]]

            #gather data on this RANK
            if RANK_CELLINDICES[RANK].size > 0:
                for i, cellindex in enumerate(RANK_CELLINDICES[RANK]):
                    if i == 0:
                        data_temp = np.zeros([RANK_CELLINDICES[RANK].size] +
                                            list(self.output[cellindex
                                                             ][measure].shape),
                                            dtype=np.float32)
                    data_temp[i, ] = self.output[cellindex][measure]

            if RANK == 0:
                #container of all output
                data = np.zeros([SAMPLESIZE] +
                                 list(self.output[cellindex][measure].shape),
                                 dtype=np.float32)

                #fill in values from this RANK
                if RANK_CELLINDICES[0].size > 0:
                    for j, k in enumerate(RANK_CELLINDICES[0]):
                        data[k, ] = data_temp[j, ]

                #iterate over all other RANKs
                for i in range(1, len(RANK_CELLINDICES)):
                    if RANK_CELLINDICES[i].size > 0:
                        #receive on RANK 0 from all other RANK
                        data_temp = np.zeros([RANK_CELLINDICES[i].size] +
                                        list(self.output[cellindex
                                                         ][measure].shape),
                                        dtype=np.float32)
                        COMM.Recv([data_temp, MPI.FLOAT], source=i, tag=13)

                        #fill in values
                        for j, k in enumerate(RANK_CELLINDICES[i]):
                            data[k, ] = data_temp[j, ]
            else:
                data = None
                if RANK_CELLINDICES[RANK].size > 0:
                    #send to RANK 0
                    COMM.Send([data_temp, MPI.FLOAT], dest=0, tag=13)

            if RANK == 0:
                #save all single-cell data to file
                fname = os.path.join(self.populations_path,
                                     '%s_%ss.h5' % (self.y, measure))
                f = h5py.File(fname, 'w')
                f.create_dataset('data', data=data, compression=4)
                f['srate'] = self.output[0]['srate']
                f.close()
                assert(os.path.isfile(fname))

                print('file %s_%ss.h5 ok' % (self.y, measure))

            COMM.Barrier()

            return data


    def collect_data(self):
        """
        Collect LFPs, CSDs and soma traces from each simulated population,
        and save to file.


        Parameters
        ----------
        None


        Returns
        -------
        None

        """
        #collect some measurements resolved per file and save to file
        for measure in ['LFP', 'CSD']:
            if measure in self.savelist:
                self.collectSingleContribs(measure)


        #calculate lfp from all cell contribs
        lfp = self.calc_signal_sum(measure='LFP')

        #calculate CSD in every lamina
        if self.calculateCSD:
            csd = self.calc_signal_sum(measure='CSD')

        if RANK == 0 and self.POPULATION_SIZE > 0:
            #saving LFPs
            if 'LFP' in self.savelist:
                fname = os.path.join(self.populations_path,
                                     self.output_file.format(self.y,
                                                             'LFP')+'.h5')
                f = h5py.File(fname, 'w')
                f['srate'] = 1E3 / self.dt_output
                f.create_dataset('data', data=lfp, compression=4)
                f.close()
                del lfp
                assert(os.path.isfile(fname))
                print('save lfp ok')


            #saving CSDs
            if 'CSD' in self.savelist and self.calculateCSD:
                fname = os.path.join(self.populations_path,
                                     self.output_file.format(self.y,
                                                             'CSD')+'.h5')
                f = h5py.File(fname, 'w')
                f['srate'] = 1E3 / self.dt_output
                f.create_dataset('data', data=csd, compression=4)
                f.close()
                del csd
                assert(os.path.isfile(fname))
                print('save CSD ok')


            #save the somatic placements:
            pop_soma_pos = np.zeros((self.POPULATION_SIZE, 3))
            keys = ['xpos', 'ypos', 'zpos']
            for i in range(self.POPULATION_SIZE):
                for j in range(3):
                    pop_soma_pos[i, j] = self.pop_soma_pos[i][keys[j]]
            fname = os.path.join(self.populations_path,
                                 self.output_file.format(self.y, 'somapos.gdf'))
            np.savetxt(fname, pop_soma_pos)
            assert(os.path.isfile(fname))
            print('save somapos ok')

            #save rotations using hdf5
            fname = os.path.join(self.populations_path,
                                    self.output_file.format(self.y, 'rotations.h5'))
            f = h5py.File(fname, 'w')
            f.create_dataset('x', (len(self.rotations),))
            f.create_dataset('y', (len(self.rotations),))
            f.create_dataset('z', (len(self.rotations),))

            for i, rot in enumerate(self.rotations):
                for key, value in list(rot.items()):
                    f[key][i] = value
            f.close()
            assert(os.path.isfile(fname))
            print('save rotations ok')


        #resync threads
        COMM.Barrier()


class Population(PopulationSuper):
    """
    Class `hybridLFPy.Population`, inherited from class `PopulationSuper`.

    This class rely on spiking times recorded in a network simulation,
    layer-resolved indegrees, synapse parameters, delay parameters, all per
    presynaptic population.


    Parameters
    ----------
    X : list of str
        Each element denote name of presynaptic populations.
    networkSim : `hybridLFPy.cachednetworks.CachedNetwork` object
        Container of network spike events resolved per population
    k_yXL : numpy.ndarray
        Num layers x num presynapse populations array specifying the
        number of incoming connections per layer and per population type.
    synParams : dict of dicts
        Synapse parameters (cf. `LFPy.Synapse` class).
        Each toplevel key denote each presynaptic population,
        bottom-level dicts are parameters passed to `LFPy.Synapse`.
    synDelayLoc : list
        Average synapse delay for each presynapse connection.
    synDelayScale : list
        Synapse delay std for each presynapse connection.
    J_yX : list of floats
        Synapse weights for connections of each presynaptic population, see
        class `LFPy.Synapse`


    Returns
    -------
    `hybridLFPy.population.Population` object


    See also
    --------
    PopulationSuper, CachedNetwork, CachedFixedSpikesNetwork,
    CachedNoiseNetwork, LFPy.Cell, LFPy.RecExtElectrode
    """

    def __init__(self,
                X = ['EX', 'IN'],
                networkSim = 'hybridLFPy.cachednetworks.CachedNetwork',
                k_yXL = [[20,  0], [20, 10]],
                synParams = {
                    'EX': {
                        'section': ['apic', 'dend'],
                        'syntype': 'AlphaISyn',
                        'tau': 0.5},
                    'IN': {
                        'section': ['dend'],
                        'syntype': 'AlphaISyn',
                        'tau': 0.5}},
                synDelayLoc = [1.5, 1.5],
                synDelayScale = [None, None],
                J_yX = [0.20680155243678455, -1.2408093146207075],
                #calculateCSD = True,
                **kwargs):
        """
        Class `hybridLFPy.Population`, inherited from class `PopulationSuper`.

        This class rely on spiking times recorded in a network simulation,
        layer-resolved indegrees, synapse parameters, delay parameters, all per
        presynaptic population.


        Parameters
        ----------
        X : list of str
            Each element denote name of presynaptic populations.
        networkSim : `hybridLFPy.cachednetworks.CachedNetwork` object
            Container of network spike events resolved per population
        k_yXL : numpy.ndarray
            Num layers x num presynapse populations array specifying the
            number of incoming connections per layer and per population type.
        synParams : dict of dicts
            Synapse parameters (cf. `LFPy.Synapse` class).
            Each toplevel key denote each presynaptic population,
            bottom-level dicts are parameters passed to `LFPy.Synapse`.
        synDelayLoc : list
            Average synapse delay for each presynapse connection.
        synDelayScale : list
            Synapse delay std for each presynapse connection.
        J_yX : list of floats
            Synapse weights for connections of each presynaptic population, see
            class `LFPy.Synapse`
        #calculateCSD : bool
        #    Flag for computing the ground-source CSD.


        Returns
        -------
        `hybridLFPy.population.Population` object


        See also
        --------
        PopulationSuper, CachedNetwork, CachedFixedSpikesNetwork,
        CachedNoiseNetwork, LFPy.Cell, LFPy.RecExtElectrode
        """
        tic = time()

        PopulationSuper.__init__(self, **kwargs)
        #set some class attributes
        self.X = X
        self.networkSim = networkSim
        self.k_yXL = np.array(k_yXL)


        #local copy of synapse parameters
        self.synParams = synParams
        self.synDelayLoc = synDelayLoc
        self.synDelayScale = synDelayScale
        self.J_yX = J_yX


        #Now loop over all cells in the population and assess
        # - number of synapses in each z-interval (from layerbounds)
        # - placement of synapses in each z-interval

        #get in this order, the
        # - postsynaptic compartment indices
        # - presynaptic cell indices
        # - synapse delays per connection
        self.synIdx = self.get_all_synIdx()
        self.SpCells = self.get_all_SpCells()
        self.synDelays = self.get_all_synDelays()

        if RANK == 0:
            print("population initialized in %.2f seconds" % (time()-tic))


    def get_all_synIdx(self):
        """
        Auxilliary function to set up class attributes containing
        synapse locations given as LFPy.Cell compartment indices

        This function takes no inputs.


        Parameters
        ----------
        None


        Returns
        -------
        synIdx : dict
            `output[cellindex][populationindex][layerindex]` numpy.ndarray of
            compartment indices.


        See also
        --------
        Population.get_synidx, Population.fetchSynIdxCell
        """
        tic = time()

        #containers for synapse idxs existing on this rank
        synIdx = {}


        #ok then, we will draw random numbers across ranks, which have to
        #be unique per cell. Now, we simply record the random state,
        #change the seed per cell, and put the original state back below.
        randomstate = np.random.get_state()

        for cellindex in self.RANK_CELLINDICES:
            #set the random seed on for each cellindex
            np.random.seed(self.POPULATIONSEED + cellindex)

            #find synapse locations for cell in parallel
            synIdx[cellindex] = self.get_synidx(cellindex)

        #reset the random number generator
        np.random.set_state(randomstate)

        if RANK == 0:
            print('found synapse locations in %.2f seconds' % (time()-tic))

        #print the number of synapses per layer from which presynapse population
        if self.verbose:
            for cellindex in self.RANK_CELLINDICES:
                for i, synidx in enumerate(synIdx[cellindex]):
                    print('to:\t%s\tcell:\t%i\tfrom:\t%s:' % (self.y,
                                                cellindex, self.X[i]),)
                    idxcount = 0
                    for idx in synidx:
                        idxcount += idx.size
                        print('\t%i' % idx.size,)
                    print('\ttotal %i' % idxcount)

        return synIdx


    def get_all_SpCells(self):
        """
        For each postsynaptic cell existing on this RANK, load or compute
        the presynaptic cell index for each synaptic connection

        This function takes no kwargs.


        Parameters
        ----------
        None


        Returns
        -------
        SpCells : dict
            `output[cellindex][populationname][layerindex]`, np.array of
            presynaptic cell indices.


        See also
        --------
        Population.fetchSpCells

        """
        tic = time()

        #container
        SpCells = {}

        #ok then, we will draw random numbers across ranks, which have to
        #be unique per cell. Now, we simply record the random state,
        #change the seed per cell, and put the original state back below.
        randomstate = np.random.get_state()

        for cellindex in self.RANK_CELLINDICES:
            #set the random seed on for each cellindex
            np.random.seed(self.POPULATIONSEED + cellindex + self.POPULATION_SIZE)

            SpCells[cellindex] = {}
            for i, X in enumerate(self.X):
                SpCells[cellindex][X] = self.fetchSpCells(
                    self.networkSim.nodes[X], self.k_yXL[:, i])

        #reset the random number generator
        np.random.set_state(randomstate)

        if RANK == 0:
            print('found presynaptic cells in %.2f seconds' % (time()-tic))

        return SpCells


    def get_all_synDelays(self):
        """
        Create and load arrays of connection delays per connection on this rank

        Get random normally distributed synaptic delays,
        returns dict of nested list of same shape as SpCells.

        Delays are rounded to dt.

        This function takes no kwargs.


        Parameters
        ----------
        None


        Returns
        -------
        dict
            output[cellindex][populationname][layerindex]`, np.array of
            delays per connection.


        See also
        --------
        numpy.random.normal

        """
        tic = time()

        #ok then, we will draw random numbers across ranks, which have to
        #be unique per cell. Now, we simply record the random state,
        #change the seed per cell, and put the original state back below.
        randomstate = np.random.get_state()

        #container
        delays = {}

        for cellindex in self.RANK_CELLINDICES:
            #set the random seed on for each cellindex
            np.random.seed(self.POPULATIONSEED + cellindex + 2*self.POPULATION_SIZE)

            delays[cellindex] = {}
            for j, X in enumerate(self.X):
                delays[cellindex][X] = []
                for i in self.k_yXL[:, j]:
                    loc = self.synDelayLoc[j]
                    loc /= self.dt
                    scale = self.synDelayScale[j]
                    if scale is not None:
                        scale /= self.dt
                        delay = np.random.normal(loc, scale, i).astype(int)
                        while np.any(delay < 1):
                            inds = delay < 1
                            delay[inds] = np.random.normal(loc, scale,
                                                        inds.sum()).astype(int)
                        delay = delay.astype(float)
                        delay *= self.dt
                    else:
                        delay = np.zeros(i) + self.synDelayLoc[j]
                    delays[cellindex][X].append(delay)

        #reset the random number generator
        np.random.set_state(randomstate)

        if RANK == 0:
            print('found delays in %.2f seconds' % (time()-tic))

        return delays


    def get_synidx(self, cellindex):
        """
        Local function, draw and return synapse locations corresponding
        to a single cell, using a random seed set as
        `POPULATIONSEED` + `cellindex`.


        Parameters
        ----------
        cellindex : int
            Index of cell object.


        Returns
        -------
        synidx : dict
            `LFPy.Cell` compartment indices


        See also
        --------
        Population.get_all_synIdx, Population.fetchSynIdxCell

        """
        #create a cell instance
        cell = self.cellsim(cellindex, return_just_cell=True)


        #local containers
        synidx = {}

        #get synaptic placements and cells from the network,
        #then set spike times,
        for i, X in enumerate(self.X):
            synidx[X] = self.fetchSynIdxCell(cell=cell,
                                             nidx=self.k_yXL[:, i],
                                             synParams=self.synParams.copy())

        return synidx


    def fetchSynIdxCell(self, cell, nidx, synParams):
        """
        Find possible synaptic placements for each cell
        As synapses are placed within layers with bounds determined by
        self.layerBoundaries, it will check this matrix accordingly, and
        use the probabilities from `self.connProbLayer to distribute.

        For each layer, the synapses are placed with probability normalized
        by membrane area of each compartment


        Parameters
        ----------
        cell : `LFPy.Cell` instance
        nidx : numpy.ndarray
            Numbers of synapses per presynaptic population X.
        synParams : which `LFPy.Synapse` parameters to use.


        Returns
        -------
        syn_idx : list
            List of arrays of synapse placements per connection.


        See also
        --------
        Population.get_all_synIdx, Population.get_synIdx, LFPy.Synapse

        """

        #segment indices in each layer is stored here, list of np.array
        syn_idx = []
        #loop over layer bounds, find synapse locations
        for i, zz in enumerate(self.layerBoundaries):
            if nidx[i] == 0:
                syn_idx.append(np.array([], dtype=int))
            else:
                syn_idx.append(cell.get_rand_idx_area_norm(
                                section=synParams['section'],
                                nidx=nidx[i],
                                z_min=zz.min(),
                                z_max=zz.max()).astype('int16'))

        return syn_idx

    def cellsim(self, cellindex, return_just_cell = False):
        """
        Do the actual simulations of LFP, using synaptic spike times from
        network simulation.


        Parameters
        ----------
        cellindex : int
            cell index between 0 and population size-1.
        return_just_cell : bool
            If True, return only the `LFPy.Cell` object
            if False, run full simulation, return None.


        Returns
        -------
        None or `LFPy.Cell` object


        See also
        --------
        hybridLFPy.csd, LFPy.Cell, LFPy.Synapse, LFPy.RecExtElectrode
        """
        tic = time()
        
        cell = LFPy.Cell(**self.cellParams)
        cell.set_pos(**self.pop_soma_pos[cellindex])
        cell.set_rotation(**self.rotations[cellindex])

        if return_just_cell:
            #with several cells, NEURON can only hold one cell at the time
            allsecnames = []
            allsec = []
            for sec in cell.allseclist:
                allsecnames.append(sec.name())
                for seg in sec:
                    allsec.append(sec.name())
            cell.allsecnames = allsecnames
            cell.allsec = allsec
            return cell
        else:
            self.insert_all_synapses(cellindex, cell)

            #electrode object where LFPs are calculated
            electrode = LFPy.RecExtElectrode(**self.electrodeParams)

            if self.calculateCSD:
                cell.tvec = np.arange(cell.totnsegs)
                cell.imem = np.eye(cell.totnsegs)
                csdcoeff = csd.true_lam_csd(cell,
                                self.populationParams['radius'], electrode.z)
                csdcoeff *= 1E6 #nA mum^-3 -> muA mm^-3 conversion
                del cell.tvec, cell.imem
                cell.simulate(electrode, dotprodcoeffs=[csdcoeff],
                              **self.simulationParams)
                cell.CSD = helpers.decimate(cell.dotprodresults[0],
                                            q=self.decimatefrac)
            else:
                cell.simulate(electrode,
                              **self.simulationParams)

            cell.LFP = helpers.decimate(electrode.LFP,
                                        q=self.decimatefrac)


            cell.x = electrode.x
            cell.y = electrode.y
            cell.z = electrode.z

            cell.electrodecoeff = electrode.electrodecoeff

            #put all necessary cell output in output dict
            for attrbt in self.savelist:
                attr = getattr(cell, attrbt)
                if type(attr) == np.ndarray:
                    self.output[cellindex][attrbt] = attr.astype('float32')
                else:
                    try:
                        self.output[cellindex][attrbt] = attr
                    except:
                        self.output[cellindex][attrbt] = str(attr)
                self.output[cellindex]['srate'] = 1E3 / self.dt_output

            print('cell %s population %s in %.2f s' % (cellindex, self.y,
                                                              time()-tic))


    def insert_all_synapses(self, cellindex, cell):

        """
        Insert all synaptic events from all presynaptic layers on
        cell object with index `cellindex`.


        Parameters
        ----------
        cellindex : int
            cell index in the population.
        cell : `LFPy.Cell` instance
            Postsynaptic target cell.


        Returns
        -------
        None


        See also
        --------
        Population.insert_synapse

        """
        for i, X in enumerate(self.X): #range(self.k_yXL.shape[1]):
            synParams = self.synParams
            synParams.update({
                'weight' : self.J_yX[i]
                })
            for j in range(len(self.synIdx[cellindex][X])):
                if self.synDelays is not None:
                    synDelays = self.synDelays[cellindex][X][j]
                else:
                    synDelays = None
                self.insert_synapses(cell = cell,
                                cellindex = cellindex,
                                synParams = synParams,
                                idx = self.synIdx[cellindex][X][j],
                                X=X,
                                SpCell = self.SpCells[cellindex][X][j],
                                synDelays = synDelays)


    def insert_synapses(self, cell, cellindex, synParams, idx = np.array([]),
                        X='EX', SpCell = np.array([]),
                        synDelays = None):
        """
        Insert synapse with `parameters`=`synparams` on cell=cell, with
        segment indexes given by `idx`. `SpCell` and `SpTimes` picked from
        Brunel network simulation


        Parameters
        ----------
        cell : `LFPy.Cell` instance
            Postsynaptic target cell.
        cellindex : int
            Index of cell in population.
        synParams : dict
            Parameters passed to `LFPy.Synapse`.
        idx : numpy.ndarray
            Postsynaptic compartment indices.
        X : str
            presynaptic population name
        SpCell : numpy.ndarray
            Presynaptic spiking cells.
        synDelays : numpy.ndarray
            Per connection specific delays.


        Returns
        -------
        None


        See also
        --------
        Population.insert_all_synapses

        """
        #Insert synapses in an iterative fashion
        try:
            spikes = self.networkSim.dbs[X].select(SpCell[:idx.size])
        except AttributeError as ae:
            raise ae, 'could not open CachedNetwork database objects'


        #apply synaptic delays
        if synDelays is not None and idx.size > 0:
            for i, delay in enumerate(synDelays):
                if spikes[i].size > 0:
                    spikes[i] += delay

        #create synapse events:
        for i in range(idx.size):
            if len(spikes[i]) == 0:
                pass
                #print 'no spike times, skipping network cell #%i' % SpCell[i]
            else:
                synParams.update({'idx' : idx[i]})
                # Create synapse(s) and setting times using class LFPy.Synapse
                synapse = LFPy.Synapse(cell, **synParams)
                #SpCell is a vector, or do not exist
                synapse.set_spike_times(spikes[i] + cell.tstartms)


    def fetchSpCells(self, nodes, numSyn):
        """
        For N (nodes count) nestSim-cells draw
        POPULATION_SIZE x NTIMES random cell indexes in
        the population in nodes and broadcast these as `SpCell`.

        The returned argument is a list with len = numSyn.size of np.arrays,
        assumes `numSyn` is a list


        Parameters
        ----------
        nodes : numpy.ndarray, dtype=int
            Node # of valid presynaptic neurons.
        numSyn : numpy.ndarray, dtype=int
            # of synapses per connection.


        Returns
        -------
        SpCells : list
            presynaptic network-neuron indices


        See also
        --------
        Population.fetch_all_SpCells
        """
        SpCell = []
        for size in numSyn:
            SpCell.append(np.random.randint(nodes.min(), nodes.max(),
                                            size=size).astype('int32'))
        return SpCell


if __name__ == '__main__':
    import doctest
    doctest.testmod()




#!/usr/bin/env python
"""
Cached networks to use with the population classes, as the only
variables being used is "nodes_ex" and "nodes_in" VERSION THAT WORKS.
"""

import numpy as np
import os
import glob
if 'DISPLAY' not in os.environ:
    import matplotlib
    matplotlib.use('Agg')
from .gdf import GDF
import matplotlib.pyplot as plt
from mpi4py import MPI


################# Initialization of MPI stuff ##################################
COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()


############## Functions #######################################################

def remove_axis_junk(ax, which=['right', 'top']):
    """
    Remove axis lines from axes object that exist in list which.

    
    Parameters
    ----------
    ax : `matplotlib.axes.AxesSubplot` object
    which : list of str
        Entries in ['right', 'top', 'bottom', 'left'].


    Returns
    -------
    None
    
    """
    for loc, spine in ax.spines.items():
        if loc in which:
            spine.set_color('none')            
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')


################ Classes #######################################################

class CachedNetwork(object):
    """
    Offline processing and storing of network spike events, used by other
    class objects in the package hybridLFPy.


    Parameters
    ----------
    simtime : float
        Simulation duration.
    dt : float,
        Simulation timestep size.
    spike_output_path : str
        Path to gdf-files with spikes.
    label : str
        Prefix of spiking gdf-files.
    ext : str
        File extension of gdf-files.
    GIDs : dict
        dictionary keys are population names and item a list with first
        GID in population and population size            
    autocollect : bool
        If True, class init will process gdf files.
    cmap : str
        Name of colormap, must be in `dir(plt.cm)`.
    
    
    Returns
    -------
    `hybridLFPy.cachednetworks.CachedNetwork` object


    See also
    --------
    CachedFixedSpikesNetwork, CachedNoiseNetwork
    """

    def __init__(self,
                 simtime = 1000.,
                 dt = 0.1,
                 spike_output_path='spike_output_path',
                 label = 'spikes',
                 ext = 'gdf',
                 GIDs={'EX' : [1, 400], 'IN' : [401, 100]},
                 autocollect=True,
                 cmap='Set1',
                 ):
        """
        Offline processing and storing of network spike events, used by other
        class objects in the package `hybridLFPy`.


        Parameters
        ----------
        simtime : float
            Simulation duration.
        dt : float
            Simulation timestep size.
        spike_output_path : str
            Path to gdf-files with spikes.
        label : str
            Prefix of spiking gdf-files.
        ext : str
            File extension of gdf-files.
        GIDs : dict
            dictionary keys are population names and item a list with first
            GID in population and population size            
        autocollect : bool
            If True, class init will process gdf files.
        cmap : str
            Name of colormap, must be in dir(plt.cm).
        
        
        Returns
        -------
        `hybridLFPy.cachednetworks.CachedNetwork` object


        See also
        --------
        CachedFixedSpikesNetwork, CachedNoiseNetwork
            
        """
        # Set some attributes
        self.simtime = simtime
        self.dt = dt
        self.spike_output_path = spike_output_path
        self.label = label
        self.ext = ext
        self.dbname = ':memory:'
        self.GIDs = GIDs
        self.X = GIDs.keys()
        self.X.sort()
        self.autocollect = autocollect

        # Create a dictionary of nodes with proper layernames
        self.nodes = {}
        for X in self.X:
            self.nodes[X] = np.arange(self.GIDs[X][1]) + self.GIDs[X][0]

        #list population sizes
        self.N_X = np.array([self.GIDs[X][1] for X in self.X])

        if self.autocollect:
            #collect the gdf files
            self.collect_gdf()


        # Specify some plot colors used for each population:
        if 'TC' in self.X:
            numcolors = len(self.X)-1
        else:
            numcolors = len(self.X)
            
        self.colors = []
        for i in range(numcolors):
            self.colors += [plt.get_cmap(cmap, numcolors)(i)]

        if 'TC' in self.X:
            self.colors += ['k']
        

    def collect_gdf(self):
        """
        Collect the gdf-files from network sim in folder `spike_output_path`
        into sqlite database, using the GDF-class.
        
        
        Parameters
        ----------
        None
        
        
        Returns
        -------
        None
        
        """
        # Resync
        COMM.Barrier()

        # Raise Exception if there are no gdf files to be read
        if len(glob.glob(os.path.join(self.spike_output_path,
                                      self.label + '*.'+ self.ext))) == 0:
            raise Exception('path to files contain no gdf-files!')

        #create in-memory databases of spikes
        if not hasattr(self, 'dbs'):
            self.dbs = {}
        
        for X in self.X:
            db = GDF(os.path.join(self.dbname),
                     debug=True, new_db=True)
            db.create(re=os.path.join(self.spike_output_path,
                                      '{0}*{1}*{2}'.format(self.label, X,
                                                           self.ext)),
                      index=True)
            self.dbs.update({
                    X : db
                })
      
        COMM.Barrier()


    def get_xy(self, xlim, fraction=1.):
        """
        Get pairs of node units and spike trains on specific time interval.
        
        
        Parameters
        ----------
        xlim : list of floats
            Spike time interval, e.g., [0., 1000.].
        fraction : float in [0, 1.]
            If less than one, sample a fraction of nodes in random order.
        
        
        Returns
        -------
        x : dict
            In `x` key-value entries are population name and neuron spike times.
        y : dict
            Where in `y` key-value entries are population name and neuron gid number.

        """
        x = {}
        y = {}

        for X, nodes in self.nodes.items():
            x[X] = np.array([])
            y[X] = np.array([])

            if fraction != 1:
                nodes = np.random.permutation(nodes)[:int(nodes.size*fraction)]
                nodes.sort()

            spiketimes = self.dbs[X].select_neurons_interval(nodes, T=xlim)
            i = 0
            for times in spiketimes:
                x[X] = np.r_[x[X], times]
                y[X] = np.r_[y[X], np.zeros(times.size) + nodes[i]]
                i += 1
                
        return x, y


    def plot_raster(self, ax, xlim, x, y, pop_names=False,
                    markersize=20., alpha=1., legend=True,
                    marker='o'):
        """
        Plot network raster plot in subplot object.
        
        
        Parameters
        ----------
        ax : `matplotlib.axes.AxesSubplot` object
            plot axes
        xlim : list
            List of floats. Spike time interval, e.g., [0., 1000.].
        x : dict
            Key-value entries are population name and neuron spike times.
        y : dict
            Key-value entries are population name and neuron gid number.
        pop_names: bool
            If True, show population names on yaxis instead of gid number.
        markersize : float
            raster plot marker size
        alpha : float in [0, 1]
            transparency of marker
        legend : bool
            Switch on axes legends.


        Returns
        -------
        None
        
        """
        yoffset = [sum(self.N_X) if X=='TC' else 0 for X in self.X]
        for i, X in enumerate(self.X):
            if y[X].size > 0:
                ax.plot(x[X], y[X]+yoffset[i], marker,
                    markersize=markersize,
                    markerfacecolor=self.colors[i],
                    markeredgecolor=self.colors[i],
                    alpha=alpha,
                    label=X, rasterized=True,
                    clip_on=True)
        
        #don't draw anything for the may-be-quiet TC population
        N_X_sum = 0
        for i, X in enumerate(self.X):
            if y[X].size > 0:
                N_X_sum += self.N_X[i]
        
        ax.axis([xlim[0], xlim[1],
                 self.GIDs[self.X[0]][0], self.GIDs[self.X[0]][0]+N_X_sum])
        ax.set_ylim(ax.get_ylim()[::-1])
        ax.set_ylabel('cell id', labelpad=0)
        ax.set_xlabel('$t$ (ms)', labelpad=0)
        if legend:
            ax.legend()
        if pop_names:
            yticks = []
            yticklabels = []
            for i, X in enumerate(self.X):
                if y[X] != []:
                    yticks.append(y[X].mean()+yoffset[i])
                    yticklabels.append(self.X[i])
            ax.set_yticks(yticks)
            ax.set_yticklabels(yticklabels)
        
        # Add some horizontal lines separating the populations
        for i, X in enumerate(self.X):
            if y[X].size > 0:
                ax.plot([xlim[0], xlim[1]],
                        [y[X].max()+yoffset[i], y[X].max()+yoffset[i]],
                        'k', lw=0.25)


    def plot_f_rate(self, ax, X, i, xlim, x, y, binsize=1, yscale='linear',
                    plottype='fill_between', show_label=False, rasterized=False):
        """
        Plot network firing rate plot in subplot object.
        
        
        Parameters
        ----------
        ax : `matplotlib.axes.AxesSubplot` object.
        X : str
            Population name.
        i : int
            Population index in class attribute `X`.
        xlim : list of floats
            Spike time interval, e.g., [0., 1000.].
        x : dict
            Key-value entries are population name and neuron spike times.
        y : dict
            Key-value entries are population name and neuron gid number.
        yscale : 'str'
            Linear, log, or symlog y-axes in rate plot.
        plottype : str
            plot type string in `['fill_between', 'bar']`
        show_label : bool
            whether or not to show labels
        

        Returns
        -------
        None
        
        """
        
        bins = np.arange(xlim[0], xlim[1]+binsize, binsize)
        (hist, bins) = np.histogram(x[X], bins=bins)
        
        if plottype == 'fill_between':
            ax.fill_between(bins[:-1], hist * 1000. / self.N_X[i],
                    color=self.colors[i], lw=0.5, label=X, rasterized=rasterized,
                    clip_on=False)
            ax.plot(bins[:-1], hist * 1000. / self.N_X[i],
                    color='k', lw=0.5, label=X, rasterized=rasterized,
                    clip_on=False)
        elif plottype == 'bar':
            ax.bar(bins[:-1], hist * 1000. / self.N_X[i],
                    color=self.colors[i], label=X, rasterized=rasterized ,
                    linewidth=0.25, width=0.9, clip_on=False)
        else:
            mssg = "plottype={} not in ['fill_between', 'bar']".format(plottype)
            raise Exception(mssg)

        remove_axis_junk(ax)

        ax.axis(ax.axis('tight'))

        ax.set_yscale(yscale)

        ax.set_xlim(xlim[0], xlim[1])
        if show_label:
            ax.text(xlim[0] + .05*(xlim[1]-xlim[0]), ax.axis()[3]*1.5, X,
                    va='center', ha='left')


    def raster_plots(self, xlim=[0, 1000], markersize=1, alpha=1., marker='o'):
        """
        Pretty plot of the spiking output of each population as raster and rate.
        
        
        Parameters
        ----------
        xlim : list
            List of floats. Spike time interval, e.g., `[0., 1000.]`.
        markersize : float
            marker size for plot, see `matplotlib.pyplot.plot`
        alpha : float
            transparency for markers, see `matplotlib.pyplot.plot`
        marker : :mod:`A valid marker style <matplotlib.markers>`
        
        
        Returns
        -------
        fig : `matplotlib.figure.Figure` object
        
        """
        x, y = self.get_xy(xlim)

        fig = plt.figure()
        fig.subplots_adjust(left=0.12, hspace=0.15)

        ax0 = fig.add_subplot(211)

        self.plot_raster(ax0, xlim, x, y, markersize=markersize, alpha=alpha,
                         marker=marker)
        remove_axis_junk(ax0)
        ax0.set_title('spike raster')
        ax0.set_xlabel("")

        nrows = len(self.X)
        bottom = np.linspace(0.1, 0.45, nrows+1)[::-1][1:]
        thickn = np.abs(np.diff(bottom))[0]*0.9


        for i, layer in enumerate(self.X):
            ax1 = fig.add_axes([0.12, bottom[i], 0.78, thickn])

            self.plot_f_rate(ax1, layer, i, xlim, x, y, )

            if i == nrows-1:
                ax1.set_xlabel('time (ms)')
            else:
                ax1.set_xticklabels([])

            if i == 4:
                ax1.set_ylabel(r'population rates ($s^{-1}$)')

            if i == 0:
                ax1.set_title(r'population firing rates ($s^{-1}$)')
              
        return fig


class CachedFixedSpikesNetwork(CachedNetwork):
    """
    Subclass of CachedNetwork.
    
    Fake nest output, where each cell in a subpopulation spike
    simultaneously, and each subpopulation is activated at times given in
    kwarg activationtimes.
    
    
    Parameters
    ----------
    activationtimes : list of floats
        Each entry set spike times of all cells in each population
    autocollect : bool
        whether or not to automatically gather gdf file output
    **kwargs : see parent class `hybridLFPy.cachednetworks.CachedNetwork`
    
    
    Returns
    -------
    `hybridLFPy.cachednetworks.CachedFixedSpikesNetwork` object


    See also
    --------
    CachedNetwork, CachedNoiseNetwork, 

    """
    def __init__(self,
                 activationtimes=[200, 300, 400, 500, 600, 700, 800, 900, 1000],
                 autocollect=False,
                 **kwargs):
        """
        Subclass of CachedNetwork
        
        Fake nest output, where each cell in a subpopulation spike
        simultaneously, and each subpopulation is activated at times given in
        kwarg activationtimes.
        
        Parameters
        ----------
        activationtimes : list
            Each entry set spike times of all cells in each population
        autocollect : bool
            whether or not to automatically gather gdf file output
        **kwargs : see parent class `hybridLFPy.cachednetworks.CachedNetwork`
        
        
        Returns
        -------
        `hybridLFPy.cachednetworks.CachedFixedSpikesNetwork` object


        See also
        --------
        CachedNetwork, CachedNoiseNetwork, 
        
        """
        
        CachedNetwork.__init__(self, autocollect=autocollect, **kwargs)

        # Set some attributes
        self.activationtimes = activationtimes
        
        if len(activationtimes) != len(self.N_X):
            raise Exception('len(activationtimes != len(self.N_X))')

        """ Create a dictionary of nodes with proper layernames
         self.nodes = {}.
        """

        if RANK == 0:
            for i, N in enumerate(self.N_X):
                nodes = self.nodes[self.X[i]]
                cell_spt = list(zip(nodes, [self.activationtimes[i]
                                  for x in range(nodes.size)]))
                cell_spt = np.array(cell_spt, dtype=[('a', int), ('b', float)])

                np.savetxt(os.path.join(self.spike_output_path,
                                        self.label + '_{}.gdf'.format(self.X[i])),
                           cell_spt, fmt=['%i', '%.1f'])

        # Resync
        COMM.barrier()

        # Collect the gdf files
        self.collect_gdf()



class CachedNoiseNetwork(CachedNetwork):
    """
    Subclass of CachedNetwork.

    Use Nest to generate N_X poisson-generators each with rate frate,
    and record every vector, and create database with spikes.

    Parameters
    ----------
    frate : list
        Rate of each layer, may be tuple (onset, rate, offset)
    autocollect : bool
        whether or not to automatically gather gdf file output
    **kwargs : see parent class `hybridLFPy.cachednetworks.CachedNetwork`


    Returns
    -------
    `hybridLFPy.cachednetworks.CachedNoiseNetwork` object


    See also
    --------
    CachedNetwork, CachedFixedSpikesNetwork 
    

    """
    def __init__(self,
                 frate=[(200., 15., 210.), 0.992, 3.027, 4.339, 5.962,
                        7.628, 8.669, 1.118, 7.859],
                 autocollect=False,
                 **kwargs):
        """
        Subclass of `CachedNetwork`.
        Use Nest to generate N_X poisson-generators each with rate frate,
        and record every vector, and create database with spikes.
    
        Parameters
        ----------
        frate : list
            Rate of each layer, may be tuple (onset, rate, offset).
        autocollect : bool
            whether or not to automatically gather gdf file output
        **kwargs : see parent class `hybridLFPy.cachednetworks.CachedNetwork`
    
    
        Returns
        -------
        `hybridLFPy.cachednetworks.CachedNoiseNetwork` object
    
    
        See also
        --------
        CachedNetwork, CachedFixedSpikesNetwork 
        
        """
        CachedNetwork.__init__(self, autocollect=autocollect, **kwargs)

        """
        Putting import nest here, avoid making `nest` a mandatory
        `hybridLFPy` dependency.
        """
        import nest


        #set some attributes:
        self.frate = frate
        if len(self.frate) != self.N_X.size:
            raise Exception('self.frate.size != self.N_X.size')

        self.spike_output_path = spike_output_path

        self.total_num_virtual_procs = SIZE

        # Reset nest kernel and set some kernel status variables, destroy old
        # nodes etc in the process
        nest.ResetKernel()

        #if dt is in powers of two, dt must be multiple of ms_per_tic
        if self.dt in 2**np.arange(-32., 0):
            nest.SetKernelStatus({
                "tics_per_ms" : 2**2 / self.dt,
                "resolution": self.dt,
                "print_time": True,
                "overwrite_files" : True,
                "total_num_virtual_procs" : self.total_num_virtual_procs,
                })
        else:
            nest.SetKernelStatus({
                "resolution": self.dt,
                "print_time": True,
                "overwrite_files" : True,
                "total_num_virtual_procs" : self.total_num_virtual_procs,
                })

        nest.SetDefaults("spike_detector", {
            'withtime' : True,
            'withgid' : True,
            'to_file' : True,
            'to_memory' : False,
        })

        # Create some populations of parrot neurons that echo the poisson noise
        self.nodes = {}
        for i, N in enumerate(self.N_X):
            self.nodes[self.X[i]] = nest.Create('parrot_neuron', N)

        if os.path.isfile(os.path.join(self.spike_output_path, self.dbname)):
            mystring = os.path.join(self.spike_output_path, self.dbname)
            print('db %s exist, will not rerun sim or collect gdf!' % mystring)
        else:
            # Create spike detector
            self.spikes = nest.Create("spike_detector", 1,
                            {'label' : os.path.join(self.spike_output_path,
                                                    self.label)})

            """ Create independent poisson spike trains with the some rate,
             but each layer population should really have different rates.
             """
            self.noise = []
            for rate in self.frate:
                if type(rate) == tuple:
                    self.noise.append(nest.Create("poisson_generator", 1,
                                                  { "start" : rate[0],
                                                    "rate" :  rate[1],
                                                    "stop" :  rate[2]}))
                else:
                    self.noise.append(nest.Create("poisson_generator", 1,
                                                  {"rate" : rate}))

            ## Connect parrots and spike detector
            for layer in self.X:
                nest.ConvergentConnect(self.nodes[layer], self.spikes,
                                       model='static_synapse')

            # Connect noise generators and nodes
            for i, layer in enumerate(self.X):
                nest.ConvergentConnect(self.noise[i], self.nodes[layer],
                                       model='static_synapse')

            # Run simulation
            nest.Simulate(self.simtime)

            # Collect the gdf files
            self.collect_gdf()

            # Nodes need to be collected in np.ndarrays:
            for key in list(self.nodes.keys()):
                self.nodes[key] = np.array(self.nodes[key])


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
Create db from gdf files to be able to then select neuron spike times.

Best use case:
1. run simulation
2. create sqlite db of spike times with indexing
3. use this db many times

Creating index for db will dominate insertions for larger set of spike times.

[TODO] Check how much slower things are if index is created at start
[TODO] Simplify block_read. sqlite probably can covert for insert.
[TODO] Error checking
[TODO] sqlite optimizations
[TODO] Create read buffers once instead of for each file
[TODO] Spike times may be storable as int32 rather than float, save space
"""
import numpy as np
import sqlite3 as sqlite
import os, glob
from time import time as now
import matplotlib.pyplot as plt


plt.rcdefaults()
plt.rcParams.update({
    'font.size' : 16,
    'axes.labelsize' : 16,
    'axes.titlesize' : 16,
    'legend.fontsize' : 14,
    'xtick.labelsize' : 16,
    'ytick.labelsize' : 16,
    'figure.subplot.wspace' : 0.3,
    'figure.subplot.hspace' : 0.3,
})

class GDF(object):
    """
    1. Read from gdf files.
    2. Create sqlite db of (neuron, spike time).
    3. Query spike times for neurons.


    Parameters
    ----------
    dbname : str
        Filename of sqlite database, see `sqlite3.connect`
    bsize : int
        Number of spike times to insert.
    new_db : bool
        New database with name dbname, will overwrite
        at a time, determines memory usage.
        
    
    Returns
    -------
    `hybridLFPy.gdf.GDF` object
    
    
    See also
    --------
    sqlite3, sqlite3.connect, sqlite3.connect.cursor
    """

    def __init__(self, dbname, bsize=int(1e6), new_db=True,
                 debug=False):
        """
        1. Read from gdf files.
        2. Create sqlite db of (neuron, spike time).
        3. Query spike times for neurons.
    
    
        Parameters
        ----------
        dbname : str
            Filename of sqlite database, see `sqlite3.connect`
        bsize : int
            Number of spike times to insert.
        new_db : bool
            New database with name dbname, will overwrite
            at a time, determines memory usage.
            
        
        Returns
        -------
        `hybridLFPy.gdf.GDF` object
        
        
        See also
        --------
        sqlite3, sqlite3.connect, sqlite3.connect.cursor
        """
        if new_db:
            try:
                os.unlink(dbname)
            except:
                print('creating new database file %s' % dbname)

        self.conn = sqlite.connect(dbname)
        self.cursor = self.conn.cursor()
        self.bsize = bsize
        self.debug = debug


    def _blockread(self, fname):
        """
        Generator yields bsize lines from gdf file.
        Hidden method.


        Parameters
        ----------
        fname : str
            Name of gdf-file.
            
        
        Yields
        ------
        list
            file contents
            
        """
        with open(fname, 'rb') as f:
            while True:
                a = []
                for i in range(self.bsize):
                    line = f.readline()
                    if not line: break
                    a.append(line.split())
                if a == []: raise StopIteration
                yield a

    def create(self, re='brunel-py-ex-*.gdf', index=True):
        """
        Create db from list of gdf file glob


        Parameters
        ----------
        re : str
            File glob to load.
        index : bool
            Create index on neurons for speed.
                    
        
        Returns
        -------
        None
        
        
        See also
        --------
        sqlite3.connect.cursor, sqlite3.connect
        
        """
        self.cursor.execute('CREATE TABLE IF NOT EXISTS spikes (neuron INT UNSIGNED, time REAL)')

        tic = now()
        for f in glob.glob(re):
            print(f)
            while True:
                try:
                    for data in self._blockread(f):
                        self.cursor.executemany('INSERT INTO spikes VALUES (?, ?)', data)
                        self.conn.commit()
                except:
                    continue
                break                

        toc = now()
        if self.debug: print('Inserts took %g seconds.' % (toc-tic))

        # Optionally, create index for speed
        if index:
            tic = now()
            self.cursor.execute('CREATE INDEX neuron_index on spikes (neuron)')
            toc = now()
            if self.debug: print('Indexed db in %g seconds.' % (toc-tic))


    def create_from_list(self, re=[], index=True):
        """
        Create db from list of arrays.


        Parameters
        ----------
        re : list
            Index of element is cell index, and element `i` an array of spike times in ms.
        index : bool
            Create index on neurons for speed.
        
        
        Returns
        -------
        None
        
        
        See also
        --------
        sqlite3.connect.cursor, sqlite3.connect
        
        """
        self.cursor.execute('CREATE TABLE IF NOT EXISTS spikes (neuron INT UNSIGNED, time REAL)')

        tic = now()
        i = 0
        for x in re:
            data = list(zip([i] * len(x), x))
            self.cursor.executemany('INSERT INTO spikes VALUES (?, ?)', data)
            i += 1
        self.conn.commit()
        toc = now()
        if self.debug: print('Inserts took %g seconds.' % (toc-tic))

        # Optionally, create index for speed
        if index:
            tic = now()
            self.cursor.execute('CREATE INDEX neuron_index on spikes (neuron)')
            toc = now()
            if self.debug: print('Indexed db in %g seconds.' % (toc-tic))


    def select(self, neurons):
        """
        Select spike trains.


        Parameters
        ----------
        neurons : numpy.ndarray or list
            Array of list of neurons.


        Returns
        -------
        list
            List of numpy.ndarray objects containing spike times.


        See also
        --------
        sqlite3.connect.cursor
        
        """
        s = []
        for neuron in neurons:
            self.cursor.execute('SELECT time FROM spikes where neuron = %d' % neuron)
            sel = self.cursor.fetchall()
            spikes = np.array(sel).flatten()
            s.append(spikes)
        return s


    def interval(self, T=[0, 1000]):
        """
        Get all spikes in a time interval T.


        Parameters
        ----------
        T : list
            Time interval.


        Returns
        -------
        s : list
            Nested list with spike times.



        See also
        --------
        sqlite3.connect.cursor
        
        """
        self.cursor.execute('SELECT * FROM spikes WHERE time BETWEEN %f AND %f' % tuple(T))
        sel = self.cursor.fetchall()
        return sel


    def select_neurons_interval(self, neurons, T=[0, 1000]):
        """
        Get all spikes from neurons in a time interval T.


        Parameters
        ----------
        neurons : list
            network neuron indices
        T : list
            Time interval.

        
        Returns
        ----------
        s : list
            Nested list with spike times.

        
        See also
        --------
        sqlite3.connect.cursor
        
        """
        s = []
        for neuron in neurons:
            self.cursor.execute('SELECT time FROM spikes WHERE time BETWEEN %f AND %f and neuron = %d'  % (T[0], T[1], neuron))
            sel = self.cursor.fetchall()
            spikes = np.array(sel).flatten()

            s.append(spikes)

        return s


    def neurons(self):
        """
        Return list of neuron indices.


        Parameters
        ----------
        None
        

        Returns
        -------
        list
            list of neuron indices
        
        
        See also
        --------
        sqlite3.connect.cursor
        
        """
        self.cursor.execute('SELECT DISTINCT neuron FROM spikes ORDER BY neuron')
        sel = self.cursor.fetchall()
        return np.array(sel).flatten()


    def num_spikes(self):
        """
        Return total number of spikes.


        Parameters
        ----------
        None
        
        
        Returns
        -------
        list

        """
        self.cursor.execute('SELECT Count(*) from spikes')
        rows = self.cursor.fetchall()[0]
        # Check against 'wc -l *ex*.gdf'
        if self.debug: print('DB has %d spikes' % rows)
        return rows


    def close(self):
        """
        Close `sqlite3.connect.cursor` and `sqlite3.connect` objects
        
        
        Parameters
        ----------
        None
        
        
        Returns
        -------
        None
        
        
        See also
        --------
        sqlite3.connect.cursor, sqlite3.connect

        """
        self.cursor.close()
        self.conn.close()


    def plotstuff(self, T=[0, 1000]):
        """
        Create a scatter plot of the contents of the database,
        with entries on the interval T.


        Parameters
        ----------
        T : list
            Time interval.
        
        
        Returns
        -------
        None
        
        
        See also
        --------
        GDF.select_neurons_interval
        """

        fig = plt.figure(figsize=(10,10))

        ax = fig.add_subplot(111)

        neurons = self.neurons()
        i = 0
        for x in self.select_neurons_interval(neurons, T):
            ax.plot(x, np.zeros(x.size) + neurons[i], 'o',
                    markersize=1, markerfacecolor='k', markeredgecolor='k',
                    alpha=0.25)
            i += 1
        ax.set_xlabel('time (ms)')
        ax.set_ylabel('neuron ID')
        ax.set_xlim(T[0], T[1])
        ax.set_ylim(neurons.min(), neurons.max())
        ax.set_title('database content on T = [%.0f, %.0f]' % (T[0], T[1]))


def test1():
    """ Need have a bunch of gdf files in current directory.
    Delete old db.
    """
    os.system('rm test.db')

    # Create db from excitatory files
    gdb = GDF('test.db', debug=True)
    gdb.create(re='brunel-py-ex-*.gdf', index=True)

    # Get spikes for neurons 1,2,3
    spikes = gdb.select([1,2,3])

    """ Wont get any spikes for these neurons
    cause they dont exist"""
    bad = gdb.select([100000,100001])

    gdb.close()

    print(spikes)
    print(bad)

if __name__ == '__main__':
    #test1()
    import doctest
    doctest.testmod()
